# -*- coding: utf-8 -*-
from __future__ import annotations

from google.protobuf import text_format
import enum
import queue
import typing

from . import _apis

# Workaround for good IDE and universal for runtime
if typing.TYPE_CHECKING:
    from _grpc.v4.protos import ydb_issue_message_pb2, ydb_operation_pb2
else:
    from ._grpc.common.protos import ydb_issue_message_pb2, ydb_operation_pb2


_TRANSPORT_STATUSES_FIRST = 401000
_CLIENT_STATUSES_FIRST = 402000


@enum.unique
class StatusCode(enum.IntEnum):
    STATUS_CODE_UNSPECIFIED = _apis.StatusIds.STATUS_CODE_UNSPECIFIED
    SUCCESS = _apis.StatusIds.SUCCESS
    BAD_REQUEST = _apis.StatusIds.BAD_REQUEST
    UNAUTHORIZED = _apis.StatusIds.UNAUTHORIZED
    INTERNAL_ERROR = _apis.StatusIds.INTERNAL_ERROR
    ABORTED = _apis.StatusIds.ABORTED
    UNAVAILABLE = _apis.StatusIds.UNAVAILABLE
    OVERLOADED = _apis.StatusIds.OVERLOADED
    SCHEME_ERROR = _apis.StatusIds.SCHEME_ERROR
    GENERIC_ERROR = _apis.StatusIds.GENERIC_ERROR
    TIMEOUT = _apis.StatusIds.TIMEOUT
    BAD_SESSION = _apis.StatusIds.BAD_SESSION
    PRECONDITION_FAILED = _apis.StatusIds.PRECONDITION_FAILED
    ALREADY_EXISTS = _apis.StatusIds.ALREADY_EXISTS
    NOT_FOUND = _apis.StatusIds.NOT_FOUND
    SESSION_EXPIRED = _apis.StatusIds.SESSION_EXPIRED
    CANCELLED = _apis.StatusIds.CANCELLED
    UNDETERMINED = _apis.StatusIds.UNDETERMINED
    UNSUPPORTED = _apis.StatusIds.UNSUPPORTED
    SESSION_BUSY = _apis.StatusIds.SESSION_BUSY
    EXTERNAL_ERROR = _apis.StatusIds.EXTERNAL_ERROR

    CONNECTION_LOST = _TRANSPORT_STATUSES_FIRST + 10
    CONNECTION_FAILURE = _TRANSPORT_STATUSES_FIRST + 20
    DEADLINE_EXCEEDED = _TRANSPORT_STATUSES_FIRST + 30
    CLIENT_INTERNAL_ERROR = _TRANSPORT_STATUSES_FIRST + 40
    UNIMPLEMENTED = _TRANSPORT_STATUSES_FIRST + 50

    UNAUTHENTICATED = _CLIENT_STATUSES_FIRST + 30
    SESSION_POOL_EMPTY = _CLIENT_STATUSES_FIRST + 40
    SESSION_POOL_CLOSED = _CLIENT_STATUSES_FIRST + 50


# TODO: convert from proto IssueMessage
class _IssueMessage:
    def __init__(self, message: str, issue_code: int, severity: int, issues) -> None:
        self.message = message
        self.issue_code = issue_code
        self.severity = severity
        self.issues = issues


class Error(Exception):
    status = None

    def __init__(self, message: str, issues: typing.Optional[typing.Iterable[_IssueMessage]] = None):
        super(Error, self).__init__(message)
        self.issues = issues
        self.message = message


class TruncatedResponseError(Error):
    status = None


class ConnectionError(Error):
    status = None


class ConnectionFailure(ConnectionError):
    status = StatusCode.CONNECTION_FAILURE


class ConnectionLost(ConnectionError):
    status = StatusCode.CONNECTION_LOST


class DeadlineExceed(Error):
    status = StatusCode.DEADLINE_EXCEEDED


class Unimplemented(ConnectionError):
    status = StatusCode.UNIMPLEMENTED


class Unauthenticated(Error):
    status = StatusCode.UNAUTHENTICATED


class BadRequest(Error):
    status = StatusCode.BAD_REQUEST


class Unauthorized(Error):
    status = StatusCode.UNAUTHORIZED


class InternalError(Error):
    status = StatusCode.INTERNAL_ERROR


class Aborted(Error):
    status = StatusCode.ABORTED


class Unavailable(Error):
    status = StatusCode.UNAVAILABLE


class Overloaded(Error):
    status = StatusCode.OVERLOADED


class SchemeError(Error):
    status = StatusCode.SCHEME_ERROR


class GenericError(Error):
    status = StatusCode.GENERIC_ERROR


class BadSession(Error):
    status = StatusCode.BAD_SESSION


class Timeout(Error):
    status = StatusCode.TIMEOUT


class PreconditionFailed(Error):
    status = StatusCode.PRECONDITION_FAILED


class NotFound(Error):
    status = StatusCode.NOT_FOUND


class AlreadyExists(Error):
    status = StatusCode.ALREADY_EXISTS


class SessionExpired(Error):
    status = StatusCode.SESSION_EXPIRED


class Cancelled(Error):
    status = StatusCode.CANCELLED


class Undetermined(Error):
    status = StatusCode.UNDETERMINED


class Unsupported(Error):
    status = StatusCode.UNSUPPORTED


class SessionBusy(Error):
    status = StatusCode.SESSION_BUSY


class ExternalError(Error):
    status = StatusCode.EXTERNAL_ERROR


class SessionPoolEmpty(Error, queue.Empty):
    status = StatusCode.SESSION_POOL_EMPTY


class SessionPoolClosed(Error):
    status = StatusCode.SESSION_POOL_CLOSED

    def __init__(self):
        super().__init__("Session pool is closed.")


class ClientInternalError(Error):
    status = StatusCode.CLIENT_INTERNAL_ERROR


class UnexpectedGrpcMessage(Error):
    def __init__(self, message: str):
        super().__init__(message)


def _format_issues(issues: typing.Iterable[ydb_issue_message_pb2.IssueMessage]) -> str:
    if not issues:
        return ""

    return " ,".join(text_format.MessageToString(issue, as_utf8=False, as_one_line=True) for issue in issues)


def _format_response(response: ydb_operation_pb2.Operation) -> str:
    fmt_issues = _format_issues(response.issues)
    return f"{fmt_issues} (server_code: {response.status})"


_success_status_codes = {StatusCode.STATUS_CODE_UNSPECIFIED, StatusCode.SUCCESS}
_server_side_error_map = {
    StatusCode.BAD_REQUEST: BadRequest,
    StatusCode.UNAUTHORIZED: Unauthorized,
    StatusCode.INTERNAL_ERROR: InternalError,
    StatusCode.ABORTED: Aborted,
    StatusCode.UNAVAILABLE: Unavailable,
    StatusCode.OVERLOADED: Overloaded,
    StatusCode.SCHEME_ERROR: SchemeError,
    StatusCode.GENERIC_ERROR: GenericError,
    StatusCode.TIMEOUT: Timeout,
    StatusCode.BAD_SESSION: BadSession,
    StatusCode.PRECONDITION_FAILED: PreconditionFailed,
    StatusCode.ALREADY_EXISTS: AlreadyExists,
    StatusCode.NOT_FOUND: NotFound,
    StatusCode.SESSION_EXPIRED: SessionExpired,
    StatusCode.CANCELLED: Cancelled,
    StatusCode.UNDETERMINED: Undetermined,
    StatusCode.UNSUPPORTED: Unsupported,
    StatusCode.SESSION_BUSY: SessionBusy,
    StatusCode.EXTERNAL_ERROR: ExternalError,
}


def _process_response(response_proto: ydb_operation_pb2.Operation) -> None:
    if response_proto.status not in _success_status_codes:
        exc_obj = _server_side_error_map.get(response_proto.status)
        raise exc_obj(_format_response(response_proto), response_proto.issues)



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
REQUIRED_FIELD_NUMBER: _ClassVar[int]
required: _descriptor.FieldDescriptor
SIZE_FIELD_NUMBER: _ClassVar[int]
size: _descriptor.FieldDescriptor
LENGTH_FIELD_NUMBER: _ClassVar[int]
length: _descriptor.FieldDescriptor
MAP_KEY_FIELD_NUMBER: _ClassVar[int]
map_key: _descriptor.FieldDescriptor
VALUE_FIELD_NUMBER: _ClassVar[int]
value: _descriptor.FieldDescriptor

class Limit(_message.Message):
    __slots__ = ("range", "lt", "le", "eq", "ge", "gt")
    class Range(_message.Message):
        __slots__ = ("min", "max")
        MIN_FIELD_NUMBER: _ClassVar[int]
        MAX_FIELD_NUMBER: _ClassVar[int]
        min: int
        max: int
        def __init__(self, min: _Optional[int] = ..., max: _Optional[int] = ...) -> None: ...
    RANGE_FIELD_NUMBER: _ClassVar[int]
    LT_FIELD_NUMBER: _ClassVar[int]
    LE_FIELD_NUMBER: _ClassVar[int]
    EQ_FIELD_NUMBER: _ClassVar[int]
    GE_FIELD_NUMBER: _ClassVar[int]
    GT_FIELD_NUMBER: _ClassVar[int]
    range: Limit.Range
    lt: int
    le: int
    eq: int
    ge: int
    gt: int
    def __init__(self, range: _Optional[_Union[Limit.Range, _Mapping]] = ..., lt: _Optional[int] = ..., le: _Optional[int] = ..., eq: _Optional[int] = ..., ge: _Optional[int] = ..., gt: _Optional[int] = ...) -> None: ...

class MapKey(_message.Message):
    __slots__ = ("length", "value")
    LENGTH_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    length: Limit
    value: str
    def __init__(self, length: _Optional[_Union[Limit, _Mapping]] = ..., value: _Optional[str] = ...) -> None: ...



import asyncio
from dataclasses import dataclass
from typing import Awaitable, Union

from ..issues import ClientInternalError

__all__ = [
    "OnCommit",
    "OnPartitionGetStartOffsetRequest",
    "OnPartitionGetStartOffsetResponse",
    "OnInitPartition",
    "OnShutdownPartition",
    "EventHandler",
]


class BaseReaderEvent:
    pass


@dataclass
class OnCommit(BaseReaderEvent):
    topic: str
    offset: int


@dataclass
class OnPartitionGetStartOffsetRequest(BaseReaderEvent):
    topic: str
    partition_id: int


@dataclass
class OnPartitionGetStartOffsetResponse:
    start_offset: int


class OnInitPartition(BaseReaderEvent):
    pass


class OnShutdownPartition:
    pass


TopicEventDispatchType = Union[OnPartitionGetStartOffsetResponse, None]


class EventHandler:
    def on_commit(self, event: OnCommit) -> Union[None, Awaitable[None]]:
        pass

    def on_partition_get_start_offset(
        self,
        event: OnPartitionGetStartOffsetRequest,
    ) -> Union[OnPartitionGetStartOffsetResponse, Awaitable[OnPartitionGetStartOffsetResponse]]:
        pass

    def on_init_partition(self, event: OnInitPartition) -> Union[None, Awaitable[None]]:
        pass

    def on_shutdown_partition(self, event: OnShutdownPartition) -> Union[None, Awaitable[None]]:
        pass

    async def _dispatch(self, event: BaseReaderEvent) -> Awaitable[TopicEventDispatchType]:
        f = None
        if isinstance(event, OnCommit):
            f = self.on_commit
        elif isinstance(event, OnPartitionGetStartOffsetRequest):
            f = self.on_partition_get_start_offset
        elif isinstance(event, OnInitPartition):
            f = self.on_init_partition
        elif isinstance(event, OnShutdownPartition):
            f = self.on_shutdown_partition
        else:
            raise ClientInternalError("Unsupported topic reader event")

        if asyncio.iscoroutinefunction(f):
            return await f(event)

        return f(event)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ydb_export_v1.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'ydb_export_v1.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v6.protos import ydb_export_pb2 as protos_dot_ydb__export__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13ydb_export_v1.proto\x12\rYdb.Export.V1\x1a\x17protos/ydb_export.proto2\xf6\x01\n\rExportService\x12K\n\nExportToYt\x12\x1d.Ydb.Export.ExportToYtRequest\x1a\x1e.Ydb.Export.ExportToYtResponse\x12K\n\nExportToS3\x12\x1d.Ydb.Export.ExportToS3Request\x1a\x1e.Ydb.Export.ExportToS3Response\x12K\n\nExportToFs\x12\x1d.Ydb.Export.ExportToFsRequest\x1a\x1e.Ydb.Export.ExportToFsResponseBQ\n\x18tech.ydb.proto.export.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Export_V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_export_v1_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030tech.ydb.proto.export.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Export_V1'
  _globals['_EXPORTSERVICE']._serialized_start=64
  _globals['_EXPORTSERVICE']._serialized_end=310
# @@protoc_insertion_point(module_scope)



from google.protobuf import duration_pb2 as _duration_pb2
from protos.annotations import validation_pb2 as _validation_pb2
from protos import ydb_issue_message_pb2 as _ydb_issue_message_pb2
from protos import ydb_operation_pb2 as _ydb_operation_pb2
from protos import ydb_query_stats_pb2 as _ydb_query_stats_pb2
from protos import ydb_status_codes_pb2 as _ydb_status_codes_pb2
from protos import ydb_formats_pb2 as _ydb_formats_pb2
from protos import ydb_value_pb2 as _ydb_value_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
EXEC_MODE_EXECUTE: ExecMode
EXEC_MODE_EXPLAIN: ExecMode
EXEC_MODE_PARSE: ExecMode
EXEC_MODE_UNSPECIFIED: ExecMode
EXEC_MODE_VALIDATE: ExecMode
EXEC_STATUS_ABORTED: ExecStatus
EXEC_STATUS_CANCELLED: ExecStatus
EXEC_STATUS_COMPLETED: ExecStatus
EXEC_STATUS_FAILED: ExecStatus
EXEC_STATUS_STARTING: ExecStatus
EXEC_STATUS_UNSPECIFIED: ExecStatus
SCHEMA_INCLUSION_MODE_ALWAYS: SchemaInclusionMode
SCHEMA_INCLUSION_MODE_FIRST_ONLY: SchemaInclusionMode
SCHEMA_INCLUSION_MODE_UNSPECIFIED: SchemaInclusionMode
STATS_MODE_BASIC: StatsMode
STATS_MODE_FULL: StatsMode
STATS_MODE_NONE: StatsMode
STATS_MODE_PROFILE: StatsMode
STATS_MODE_UNSPECIFIED: StatsMode
SYNTAX_PG: Syntax
SYNTAX_UNSPECIFIED: Syntax
SYNTAX_YQL_V1: Syntax

class AttachSessionRequest(_message.Message):
    __slots__ = ["session_id"]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    def __init__(self, session_id: _Optional[str] = ...) -> None: ...

class BeginTransactionRequest(_message.Message):
    __slots__ = ["session_id", "tx_settings"]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    TX_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    tx_settings: TransactionSettings
    def __init__(self, session_id: _Optional[str] = ..., tx_settings: _Optional[_Union[TransactionSettings, _Mapping]] = ...) -> None: ...

class BeginTransactionResponse(_message.Message):
    __slots__ = ["issues", "status", "tx_meta"]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    TX_META_FIELD_NUMBER: _ClassVar[int]
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    tx_meta: TransactionMeta
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ..., tx_meta: _Optional[_Union[TransactionMeta, _Mapping]] = ...) -> None: ...

class CommitTransactionRequest(_message.Message):
    __slots__ = ["session_id", "tx_id"]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    TX_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    tx_id: str
    def __init__(self, session_id: _Optional[str] = ..., tx_id: _Optional[str] = ...) -> None: ...

class CommitTransactionResponse(_message.Message):
    __slots__ = ["issues", "status"]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ...) -> None: ...

class CreateSessionRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class CreateSessionResponse(_message.Message):
    __slots__ = ["issues", "node_id", "session_id", "status"]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    node_id: int
    session_id: str
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ..., session_id: _Optional[str] = ..., node_id: _Optional[int] = ...) -> None: ...

class DeleteSessionRequest(_message.Message):
    __slots__ = ["session_id"]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    def __init__(self, session_id: _Optional[str] = ...) -> None: ...

class DeleteSessionResponse(_message.Message):
    __slots__ = ["issues", "status"]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ...) -> None: ...

class ExecuteQueryRequest(_message.Message):
    __slots__ = ["arrow_format_settings", "concurrent_result_sets", "exec_mode", "parameters", "pool_id", "query_content", "response_part_limit_bytes", "result_set_format", "schema_inclusion_mode", "session_id", "stats_mode", "stats_period_ms", "tx_control"]
    class ParametersEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _ydb_value_pb2.TypedValue
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_ydb_value_pb2.TypedValue, _Mapping]] = ...) -> None: ...
    ARROW_FORMAT_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    CONCURRENT_RESULT_SETS_FIELD_NUMBER: _ClassVar[int]
    EXEC_MODE_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    POOL_ID_FIELD_NUMBER: _ClassVar[int]
    QUERY_CONTENT_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_PART_LIMIT_BYTES_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_FORMAT_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_INCLUSION_MODE_FIELD_NUMBER: _ClassVar[int]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    STATS_MODE_FIELD_NUMBER: _ClassVar[int]
    STATS_PERIOD_MS_FIELD_NUMBER: _ClassVar[int]
    TX_CONTROL_FIELD_NUMBER: _ClassVar[int]
    arrow_format_settings: _ydb_formats_pb2.ArrowFormatSettings
    concurrent_result_sets: bool
    exec_mode: ExecMode
    parameters: _containers.MessageMap[str, _ydb_value_pb2.TypedValue]
    pool_id: str
    query_content: QueryContent
    response_part_limit_bytes: int
    result_set_format: _ydb_value_pb2.ResultSet.Format
    schema_inclusion_mode: SchemaInclusionMode
    session_id: str
    stats_mode: StatsMode
    stats_period_ms: int
    tx_control: TransactionControl
    def __init__(self, session_id: _Optional[str] = ..., exec_mode: _Optional[_Union[ExecMode, str]] = ..., tx_control: _Optional[_Union[TransactionControl, _Mapping]] = ..., query_content: _Optional[_Union[QueryContent, _Mapping]] = ..., parameters: _Optional[_Mapping[str, _ydb_value_pb2.TypedValue]] = ..., stats_mode: _Optional[_Union[StatsMode, str]] = ..., concurrent_result_sets: bool = ..., response_part_limit_bytes: _Optional[int] = ..., pool_id: _Optional[str] = ..., stats_period_ms: _Optional[int] = ..., schema_inclusion_mode: _Optional[_Union[SchemaInclusionMode, str]] = ..., result_set_format: _Optional[_Union[_ydb_value_pb2.ResultSet.Format, str]] = ..., arrow_format_settings: _Optional[_Union[_ydb_formats_pb2.ArrowFormatSettings, _Mapping]] = ...) -> None: ...

class ExecuteQueryResponsePart(_message.Message):
    __slots__ = ["exec_stats", "issues", "result_set", "result_set_index", "status", "tx_meta"]
    EXEC_STATS_FIELD_NUMBER: _ClassVar[int]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_INDEX_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    TX_META_FIELD_NUMBER: _ClassVar[int]
    exec_stats: _ydb_query_stats_pb2.QueryStats
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    result_set: _ydb_value_pb2.ResultSet
    result_set_index: int
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    tx_meta: TransactionMeta
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ..., result_set_index: _Optional[int] = ..., result_set: _Optional[_Union[_ydb_value_pb2.ResultSet, _Mapping]] = ..., exec_stats: _Optional[_Union[_ydb_query_stats_pb2.QueryStats, _Mapping]] = ..., tx_meta: _Optional[_Union[TransactionMeta, _Mapping]] = ...) -> None: ...

class ExecuteScriptMetadata(_message.Message):
    __slots__ = ["exec_mode", "exec_stats", "exec_status", "execution_id", "result_sets_meta", "script_content"]
    EXECUTION_ID_FIELD_NUMBER: _ClassVar[int]
    EXEC_MODE_FIELD_NUMBER: _ClassVar[int]
    EXEC_STATS_FIELD_NUMBER: _ClassVar[int]
    EXEC_STATUS_FIELD_NUMBER: _ClassVar[int]
    RESULT_SETS_META_FIELD_NUMBER: _ClassVar[int]
    SCRIPT_CONTENT_FIELD_NUMBER: _ClassVar[int]
    exec_mode: ExecMode
    exec_stats: _ydb_query_stats_pb2.QueryStats
    exec_status: ExecStatus
    execution_id: str
    result_sets_meta: _containers.RepeatedCompositeFieldContainer[ResultSetMeta]
    script_content: QueryContent
    def __init__(self, execution_id: _Optional[str] = ..., exec_status: _Optional[_Union[ExecStatus, str]] = ..., script_content: _Optional[_Union[QueryContent, _Mapping]] = ..., result_sets_meta: _Optional[_Iterable[_Union[ResultSetMeta, _Mapping]]] = ..., exec_mode: _Optional[_Union[ExecMode, str]] = ..., exec_stats: _Optional[_Union[_ydb_query_stats_pb2.QueryStats, _Mapping]] = ...) -> None: ...

class ExecuteScriptRequest(_message.Message):
    __slots__ = ["exec_mode", "operation_params", "parameters", "pool_id", "results_ttl", "script_content", "stats_mode"]
    class ParametersEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _ydb_value_pb2.TypedValue
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_ydb_value_pb2.TypedValue, _Mapping]] = ...) -> None: ...
    EXEC_MODE_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    POOL_ID_FIELD_NUMBER: _ClassVar[int]
    RESULTS_TTL_FIELD_NUMBER: _ClassVar[int]
    SCRIPT_CONTENT_FIELD_NUMBER: _ClassVar[int]
    STATS_MODE_FIELD_NUMBER: _ClassVar[int]
    exec_mode: ExecMode
    operation_params: _ydb_operation_pb2.OperationParams
    parameters: _containers.MessageMap[str, _ydb_value_pb2.TypedValue]
    pool_id: str
    results_ttl: _duration_pb2.Duration
    script_content: QueryContent
    stats_mode: StatsMode
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., exec_mode: _Optional[_Union[ExecMode, str]] = ..., script_content: _Optional[_Union[QueryContent, _Mapping]] = ..., parameters: _Optional[_Mapping[str, _ydb_value_pb2.TypedValue]] = ..., stats_mode: _Optional[_Union[StatsMode, str]] = ..., results_ttl: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., pool_id: _Optional[str] = ...) -> None: ...

class FetchScriptResultsRequest(_message.Message):
    __slots__ = ["fetch_token", "operation_id", "result_set_index", "rows_limit"]
    FETCH_TOKEN_FIELD_NUMBER: _ClassVar[int]
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_INDEX_FIELD_NUMBER: _ClassVar[int]
    ROWS_LIMIT_FIELD_NUMBER: _ClassVar[int]
    fetch_token: str
    operation_id: str
    result_set_index: int
    rows_limit: int
    def __init__(self, operation_id: _Optional[str] = ..., result_set_index: _Optional[int] = ..., fetch_token: _Optional[str] = ..., rows_limit: _Optional[int] = ...) -> None: ...

class FetchScriptResultsResponse(_message.Message):
    __slots__ = ["issues", "next_fetch_token", "result_set", "result_set_index", "status"]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    NEXT_FETCH_TOKEN_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_FIELD_NUMBER: _ClassVar[int]
    RESULT_SET_INDEX_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    next_fetch_token: str
    result_set: _ydb_value_pb2.ResultSet
    result_set_index: int
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ..., result_set_index: _Optional[int] = ..., result_set: _Optional[_Union[_ydb_value_pb2.ResultSet, _Mapping]] = ..., next_fetch_token: _Optional[str] = ...) -> None: ...

class OnlineModeSettings(_message.Message):
    __slots__ = ["allow_inconsistent_reads"]
    ALLOW_INCONSISTENT_READS_FIELD_NUMBER: _ClassVar[int]
    allow_inconsistent_reads: bool
    def __init__(self, allow_inconsistent_reads: bool = ...) -> None: ...

class QueryContent(_message.Message):
    __slots__ = ["syntax", "text"]
    SYNTAX_FIELD_NUMBER: _ClassVar[int]
    TEXT_FIELD_NUMBER: _ClassVar[int]
    syntax: Syntax
    text: str
    def __init__(self, syntax: _Optional[_Union[Syntax, str]] = ..., text: _Optional[str] = ...) -> None: ...

class ResultSetMeta(_message.Message):
    __slots__ = ["columns"]
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    columns: _containers.RepeatedCompositeFieldContainer[_ydb_value_pb2.Column]
    def __init__(self, columns: _Optional[_Iterable[_Union[_ydb_value_pb2.Column, _Mapping]]] = ...) -> None: ...

class RollbackTransactionRequest(_message.Message):
    __slots__ = ["session_id", "tx_id"]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    TX_ID_FIELD_NUMBER: _ClassVar[int]
    session_id: str
    tx_id: str
    def __init__(self, session_id: _Optional[str] = ..., tx_id: _Optional[str] = ...) -> None: ...

class RollbackTransactionResponse(_message.Message):
    __slots__ = ["issues", "status"]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ...) -> None: ...

class Script(_message.Message):
    __slots__ = ["script_content"]
    SCRIPT_CONTENT_FIELD_NUMBER: _ClassVar[int]
    script_content: QueryContent
    def __init__(self, script_content: _Optional[_Union[QueryContent, _Mapping]] = ...) -> None: ...

class SerializableModeSettings(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class SessionState(_message.Message):
    __slots__ = ["issues", "status"]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ...) -> None: ...

class SnapshotModeSettings(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class SnapshotRWModeSettings(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class StaleModeSettings(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class TransactionControl(_message.Message):
    __slots__ = ["begin_tx", "commit_tx", "tx_id"]
    BEGIN_TX_FIELD_NUMBER: _ClassVar[int]
    COMMIT_TX_FIELD_NUMBER: _ClassVar[int]
    TX_ID_FIELD_NUMBER: _ClassVar[int]
    begin_tx: TransactionSettings
    commit_tx: bool
    tx_id: str
    def __init__(self, tx_id: _Optional[str] = ..., begin_tx: _Optional[_Union[TransactionSettings, _Mapping]] = ..., commit_tx: bool = ...) -> None: ...

class TransactionMeta(_message.Message):
    __slots__ = ["id"]
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class TransactionSettings(_message.Message):
    __slots__ = ["online_read_only", "serializable_read_write", "snapshot_read_only", "snapshot_read_write", "stale_read_only"]
    ONLINE_READ_ONLY_FIELD_NUMBER: _ClassVar[int]
    SERIALIZABLE_READ_WRITE_FIELD_NUMBER: _ClassVar[int]
    SNAPSHOT_READ_ONLY_FIELD_NUMBER: _ClassVar[int]
    SNAPSHOT_READ_WRITE_FIELD_NUMBER: _ClassVar[int]
    STALE_READ_ONLY_FIELD_NUMBER: _ClassVar[int]
    online_read_only: OnlineModeSettings
    serializable_read_write: SerializableModeSettings
    snapshot_read_only: SnapshotModeSettings
    snapshot_read_write: SnapshotRWModeSettings
    stale_read_only: StaleModeSettings
    def __init__(self, serializable_read_write: _Optional[_Union[SerializableModeSettings, _Mapping]] = ..., online_read_only: _Optional[_Union[OnlineModeSettings, _Mapping]] = ..., stale_read_only: _Optional[_Union[StaleModeSettings, _Mapping]] = ..., snapshot_read_only: _Optional[_Union[SnapshotModeSettings, _Mapping]] = ..., snapshot_read_write: _Optional[_Union[SnapshotRWModeSettings, _Mapping]] = ...) -> None: ...

class Syntax(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class ExecMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class StatsMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class SchemaInclusionMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class ExecStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from protos import ydb_operation_pb2 as _ydb_operation_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class LoginRequest(_message.Message):
    __slots__ = ["operation_params", "password", "user"]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PASSWORD_FIELD_NUMBER: _ClassVar[int]
    USER_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    password: str
    user: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., user: _Optional[str] = ..., password: _Optional[str] = ...) -> None: ...

class LoginResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class LoginResult(_message.Message):
    __slots__ = ["token"]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    token: str
    def __init__(self, token: _Optional[str] = ...) -> None: ...



from typing import Optional

from ..._grpc.grpcwrapper.ydb_coordination import (
    CreateNodeRequest,
    DescribeNodeRequest,
    AlterNodeRequest,
    DropNodeRequest,
)
from ..._grpc.grpcwrapper.ydb_coordination_public_types import NodeConfig
from ...coordination.base import BaseCoordinationClient
from .session import CoordinationSession


class CoordinationClient(BaseCoordinationClient):
    async def create_node(self, path: str, config: Optional[NodeConfig] = None, settings=None):
        self._log_experimental_api()

        return await self._call_create(
            CreateNodeRequest(path=path, config=config).to_proto(),
            settings=settings,
        )

    async def describe_node(self, path: str, settings=None) -> NodeConfig:
        self._log_experimental_api()

        return await self._call_describe(
            DescribeNodeRequest(path=path).to_proto(),
            settings=settings,
        )

    async def alter_node(self, path: str, new_config: NodeConfig, settings=None):
        self._log_experimental_api()

        return await self._call_alter(
            AlterNodeRequest(path=path, config=new_config).to_proto(),
            settings=settings,
        )

    async def delete_node(self, path: str, settings=None):
        self._log_experimental_api()

        return await self._call_delete(
            DropNodeRequest(path=path).to_proto(),
            settings=settings,
        )

    def session(self, path: str) -> CoordinationSession:
        self._log_experimental_api()

        return CoordinationSession(self._driver, path)



from ... import StatusCode, issues

from ..._grpc.grpcwrapper.ydb_coordination import (
    AcquireSemaphore,
    ReleaseSemaphore,
    UpdateSemaphore,
    DescribeSemaphore,
    CreateSemaphore,
)
from ..._grpc.grpcwrapper.ydb_coordination_public_types import (
    DescribeLockResult,
)


class CoordinationSemaphore:
    def __init__(self, session, name: str, limit: int):
        self._session = session
        self._name = name

        self._limit = limit
        self._timeout_millis = session._timeout_millis

    async def acquire(self, count: int = 1):
        await self._create_if_not_exists()
        resp = await self._try_acquire(count)

        if resp.status != StatusCode.SUCCESS:
            raise issues.Error(f"Failed to acquire lock {self._name}: {resp.status}")

        return self

    async def release(self):
        req = ReleaseSemaphore(
            req_id=await self._session.next_req_id(),
            name=self._name,
        )
        try:
            await self._session._reconnector.send_and_wait(req)
        except Exception:
            pass

    async def describe(self) -> DescribeLockResult:
        req = DescribeSemaphore(
            req_id=await self._session.next_req_id(),
            name=self._name,
            include_owners=True,
            include_waiters=True,
            watch_data=False,
            watch_owners=False,
        )
        resp = await self._session._reconnector.send_and_wait(req)
        return DescribeLockResult.from_proto(resp)

    async def update(self, new_data: bytes) -> None:
        req = UpdateSemaphore(
            req_id=await self._session.next_req_id(),
            name=self._name,
            data=new_data,
        )
        resp = await self._session._reconnector.send_and_wait(req)

        if resp.status != StatusCode.SUCCESS:
            raise issues.Error(f"Failed to update lock {self._name}: {resp.status}")

    async def close(self):
        await self.release()

    async def __aenter__(self):
        await self.acquire()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.release()

    async def _try_acquire(self, count: int):
        req = AcquireSemaphore(
            req_id=await self._session.next_req_id(),
            name=self._name,
            count=count,
            ephemeral=False,
            timeout_millis=self._timeout_millis,
        )
        return await self._session._reconnector.send_and_wait(req)

    async def _create_if_not_exists(self):
        req = CreateSemaphore(
            req_id=await self._session.next_req_id(),
            name=self._name,
            limit=self._limit,
            data=b"",
        )
        resp = await self._session._reconnector.send_and_wait(req)

        if resp.status not in (
            StatusCode.SUCCESS,
            StatusCode.ALREADY_EXISTS,
        ):
            raise issues.Error(f"Failed to create lock {self._name}: {resp.status}")



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/ydb_discovery.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v5.protos import ydb_operation_pb2 as protos_dot_ydb__operation__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1aprotos/ydb_discovery.proto\x12\rYdb.Discovery\x1a\x1aprotos/ydb_operation.proto\"9\n\x14ListEndpointsRequest\x12\x10\n\x08\x64\x61tabase\x18\x01 \x01(\t\x12\x0f\n\x07service\x18\x02 \x03(\t\"\xc3\x01\n\x0c\x45ndpointInfo\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\r\x12\x13\n\x0bload_factor\x18\x03 \x01(\x02\x12\x0b\n\x03ssl\x18\x04 \x01(\x08\x12\x0f\n\x07service\x18\x05 \x03(\t\x12\x10\n\x08location\x18\x06 \x01(\t\x12\x0f\n\x07node_id\x18\x07 \x01(\r\x12\r\n\x05ip_v4\x18\x08 \x03(\t\x12\r\n\x05ip_v6\x18\t \x03(\t\x12 \n\x18ssl_target_name_override\x18\n \x01(\t\"\\\n\x13ListEndpointsResult\x12.\n\tendpoints\x18\x01 \x03(\x0b\x32\x1b.Ydb.Discovery.EndpointInfo\x12\x15\n\rself_location\x18\x02 \x01(\t\"E\n\x15ListEndpointsResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\'\n\rWhoAmIRequest\x12\x16\n\x0einclude_groups\x18\x01 \x01(\x08\",\n\x0cWhoAmIResult\x12\x0c\n\x04user\x18\x01 \x01(\t\x12\x0e\n\x06groups\x18\x02 \x03(\t\">\n\x0eWhoAmIResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\xe0\x02\n\x0cNodeLocation\x12 \n\x0f\x64\x61ta_center_num\x18\x01 \x01(\rB\x02\x18\x01H\x00\x88\x01\x01\x12\x19\n\x08room_num\x18\x02 \x01(\rB\x02\x18\x01H\x01\x88\x01\x01\x12\x19\n\x08rack_num\x18\x03 \x01(\rB\x02\x18\x01H\x02\x88\x01\x01\x12\x19\n\x08\x62ody_num\x18\x04 \x01(\rB\x02\x18\x01H\x03\x88\x01\x01\x12\x17\n\x04\x62ody\x18\x94\x91\x06 \x01(\rB\x02\x18\x01H\x04\x88\x01\x01\x12\x18\n\x0b\x64\x61ta_center\x18\n \x01(\tH\x05\x88\x01\x01\x12\x13\n\x06module\x18\x14 \x01(\tH\x06\x88\x01\x01\x12\x11\n\x04rack\x18\x1e \x01(\tH\x07\x88\x01\x01\x12\x11\n\x04unit\x18( \x01(\tH\x08\x88\x01\x01\x42\x12\n\x10_data_center_numB\x0b\n\t_room_numB\x0b\n\t_rack_numB\x0b\n\t_body_numB\x07\n\x05_bodyB\x0e\n\x0c_data_centerB\t\n\x07_moduleB\x07\n\x05_rackB\x07\n\x05_unitBl\n\x18tech.ydb.proto.discoveryB\x0f\x44iscoveryProtosZ<github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Discovery\xf8\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protos.ydb_discovery_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030tech.ydb.proto.discoveryB\017DiscoveryProtosZ<github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Discovery\370\001\001'
  _NODELOCATION.fields_by_name['data_center_num']._options = None
  _NODELOCATION.fields_by_name['data_center_num']._serialized_options = b'\030\001'
  _NODELOCATION.fields_by_name['room_num']._options = None
  _NODELOCATION.fields_by_name['room_num']._serialized_options = b'\030\001'
  _NODELOCATION.fields_by_name['rack_num']._options = None
  _NODELOCATION.fields_by_name['rack_num']._serialized_options = b'\030\001'
  _NODELOCATION.fields_by_name['body_num']._options = None
  _NODELOCATION.fields_by_name['body_num']._serialized_options = b'\030\001'
  _NODELOCATION.fields_by_name['body']._options = None
  _NODELOCATION.fields_by_name['body']._serialized_options = b'\030\001'
  _LISTENDPOINTSREQUEST._serialized_start=73
  _LISTENDPOINTSREQUEST._serialized_end=130
  _ENDPOINTINFO._serialized_start=133
  _ENDPOINTINFO._serialized_end=328
  _LISTENDPOINTSRESULT._serialized_start=330
  _LISTENDPOINTSRESULT._serialized_end=422
  _LISTENDPOINTSRESPONSE._serialized_start=424
  _LISTENDPOINTSRESPONSE._serialized_end=493
  _WHOAMIREQUEST._serialized_start=495
  _WHOAMIREQUEST._serialized_end=534
  _WHOAMIRESULT._serialized_start=536
  _WHOAMIRESULT._serialized_end=580
  _WHOAMIRESPONSE._serialized_start=582
  _WHOAMIRESPONSE._serialized_end=644
  _NODELOCATION._serialized_start=647
  _NODELOCATION._serialized_end=999
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ydb_operation_v1.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'ydb_operation_v1.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v6.protos import ydb_operation_pb2 as protos_dot_ydb__operation__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16ydb_operation_v1.proto\x12\x10Ydb.Operation.V1\x1a\x1aprotos/ydb_operation.proto2\x96\x03\n\x10OperationService\x12Y\n\x0cGetOperation\x12#.Ydb.Operations.GetOperationRequest\x1a$.Ydb.Operations.GetOperationResponse\x12\x62\n\x0f\x43\x61ncelOperation\x12&.Ydb.Operations.CancelOperationRequest\x1a\'.Ydb.Operations.CancelOperationResponse\x12\x62\n\x0f\x46orgetOperation\x12&.Ydb.Operations.ForgetOperationRequest\x1a\'.Ydb.Operations.ForgetOperationResponse\x12_\n\x0eListOperations\x12%.Ydb.Operations.ListOperationsRequest\x1a&.Ydb.Operations.ListOperationsResponseBW\n\x1btech.ydb.proto.operation.v1Z8github.com/ydb-platform/ydb-go-genproto/Ydb_Operation_V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_operation_v1_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033tech.ydb.proto.operation.v1Z8github.com/ydb-platform/ydb-go-genproto/Ydb_Operation_V1'
  _globals['_OPERATIONSERVICE']._serialized_start=73
  _globals['_OPERATIONSERVICE']._serialized_end=479
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ydb._grpc.v5.draft.protos import ydb_dynamic_config_pb2 as draft_dot_protos_dot_ydb__dynamic__config__pb2


class DynamicConfigServiceStub(object):
    """Dynamic Config is the configuration applied to dynamic nodes without static configuration
    and is utilized by certain services, such as logging, which can update their settings
    seamlessly without requiring any restarts.

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SetConfig = channel.unary_unary(
                '/Ydb.DynamicConfig.V1.DynamicConfigService/SetConfig',
                request_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.SetConfigRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.SetConfigResponse.FromString,
                )
        self.ReplaceConfig = channel.unary_unary(
                '/Ydb.DynamicConfig.V1.DynamicConfigService/ReplaceConfig',
                request_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ReplaceConfigRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ReplaceConfigResponse.FromString,
                )
        self.GetMetadata = channel.unary_unary(
                '/Ydb.DynamicConfig.V1.DynamicConfigService/GetMetadata',
                request_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetMetadataRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetMetadataResponse.FromString,
                )
        self.GetConfig = channel.unary_unary(
                '/Ydb.DynamicConfig.V1.DynamicConfigService/GetConfig',
                request_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetConfigRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetConfigResponse.FromString,
                )
        self.DropConfig = channel.unary_unary(
                '/Ydb.DynamicConfig.V1.DynamicConfigService/DropConfig',
                request_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.DropConfigRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.DropConfigResponse.FromString,
                )
        self.AddVolatileConfig = channel.unary_unary(
                '/Ydb.DynamicConfig.V1.DynamicConfigService/AddVolatileConfig',
                request_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.AddVolatileConfigRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.AddVolatileConfigResponse.FromString,
                )
        self.RemoveVolatileConfig = channel.unary_unary(
                '/Ydb.DynamicConfig.V1.DynamicConfigService/RemoveVolatileConfig',
                request_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.RemoveVolatileConfigRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.RemoveVolatileConfigResponse.FromString,
                )
        self.GetNodeLabels = channel.unary_unary(
                '/Ydb.DynamicConfig.V1.DynamicConfigService/GetNodeLabels',
                request_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetNodeLabelsRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetNodeLabelsResponse.FromString,
                )
        self.ResolveConfig = channel.unary_unary(
                '/Ydb.DynamicConfig.V1.DynamicConfigService/ResolveConfig',
                request_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveConfigRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveConfigResponse.FromString,
                )
        self.ResolveAllConfig = channel.unary_unary(
                '/Ydb.DynamicConfig.V1.DynamicConfigService/ResolveAllConfig',
                request_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveAllConfigRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveAllConfigResponse.FromString,
                )


class DynamicConfigServiceServicer(object):
    """Dynamic Config is the configuration applied to dynamic nodes without static configuration
    and is utilized by certain services, such as logging, which can update their settings
    seamlessly without requiring any restarts.

    """

    def SetConfig(self, request, context):
        """Apply new version of config.
        Overwrites previous version.
        This call is idempotent:
        - Two subsequent identical calls will return success,
        - After applying next version another identical call will result in an error.
        The field version within the YAML in the request must strictly be set to the current version increment by one.
        The field cluster within the YAML should be identical to the one configured on the node used by the console tablet.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReplaceConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetMetadata(self, request, context):
        """Get current configs metadata.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetConfig(self, request, context):
        """Get current configs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DropConfig(self, request, context):
        """Drop current config.
        This call is idempotent.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddVolatileConfig(self, request, context):
        """Add volatile config.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemoveVolatileConfig(self, request, context):
        """Remove volatile config.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetNodeLabels(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResolveConfig(self, request, context):
        """Resolve config for particular labels.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResolveAllConfig(self, request, context):
        """Resolve config for all possible labels combinations.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DynamicConfigServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SetConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.SetConfig,
                    request_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.SetConfigRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.SetConfigResponse.SerializeToString,
            ),
            'ReplaceConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.ReplaceConfig,
                    request_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ReplaceConfigRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ReplaceConfigResponse.SerializeToString,
            ),
            'GetMetadata': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMetadata,
                    request_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetMetadataRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetMetadataResponse.SerializeToString,
            ),
            'GetConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.GetConfig,
                    request_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetConfigRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetConfigResponse.SerializeToString,
            ),
            'DropConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.DropConfig,
                    request_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.DropConfigRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.DropConfigResponse.SerializeToString,
            ),
            'AddVolatileConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.AddVolatileConfig,
                    request_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.AddVolatileConfigRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.AddVolatileConfigResponse.SerializeToString,
            ),
            'RemoveVolatileConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.RemoveVolatileConfig,
                    request_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.RemoveVolatileConfigRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.RemoveVolatileConfigResponse.SerializeToString,
            ),
            'GetNodeLabels': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNodeLabels,
                    request_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetNodeLabelsRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.GetNodeLabelsResponse.SerializeToString,
            ),
            'ResolveConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.ResolveConfig,
                    request_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveConfigRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveConfigResponse.SerializeToString,
            ),
            'ResolveAllConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.ResolveAllConfig,
                    request_deserializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveAllConfigRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveAllConfigResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Ydb.DynamicConfig.V1.DynamicConfigService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class DynamicConfigService(object):
    """Dynamic Config is the configuration applied to dynamic nodes without static configuration
    and is utilized by certain services, such as logging, which can update their settings
    seamlessly without requiring any restarts.

    """

    @staticmethod
    def SetConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.DynamicConfig.V1.DynamicConfigService/SetConfig',
            draft_dot_protos_dot_ydb__dynamic__config__pb2.SetConfigRequest.SerializeToString,
            draft_dot_protos_dot_ydb__dynamic__config__pb2.SetConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ReplaceConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.DynamicConfig.V1.DynamicConfigService/ReplaceConfig',
            draft_dot_protos_dot_ydb__dynamic__config__pb2.ReplaceConfigRequest.SerializeToString,
            draft_dot_protos_dot_ydb__dynamic__config__pb2.ReplaceConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetMetadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.DynamicConfig.V1.DynamicConfigService/GetMetadata',
            draft_dot_protos_dot_ydb__dynamic__config__pb2.GetMetadataRequest.SerializeToString,
            draft_dot_protos_dot_ydb__dynamic__config__pb2.GetMetadataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.DynamicConfig.V1.DynamicConfigService/GetConfig',
            draft_dot_protos_dot_ydb__dynamic__config__pb2.GetConfigRequest.SerializeToString,
            draft_dot_protos_dot_ydb__dynamic__config__pb2.GetConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DropConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.DynamicConfig.V1.DynamicConfigService/DropConfig',
            draft_dot_protos_dot_ydb__dynamic__config__pb2.DropConfigRequest.SerializeToString,
            draft_dot_protos_dot_ydb__dynamic__config__pb2.DropConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddVolatileConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.DynamicConfig.V1.DynamicConfigService/AddVolatileConfig',
            draft_dot_protos_dot_ydb__dynamic__config__pb2.AddVolatileConfigRequest.SerializeToString,
            draft_dot_protos_dot_ydb__dynamic__config__pb2.AddVolatileConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RemoveVolatileConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.DynamicConfig.V1.DynamicConfigService/RemoveVolatileConfig',
            draft_dot_protos_dot_ydb__dynamic__config__pb2.RemoveVolatileConfigRequest.SerializeToString,
            draft_dot_protos_dot_ydb__dynamic__config__pb2.RemoveVolatileConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetNodeLabels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.DynamicConfig.V1.DynamicConfigService/GetNodeLabels',
            draft_dot_protos_dot_ydb__dynamic__config__pb2.GetNodeLabelsRequest.SerializeToString,
            draft_dot_protos_dot_ydb__dynamic__config__pb2.GetNodeLabelsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ResolveConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.DynamicConfig.V1.DynamicConfigService/ResolveConfig',
            draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveConfigRequest.SerializeToString,
            draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ResolveAllConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.DynamicConfig.V1.DynamicConfigService/ResolveAllConfig',
            draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveAllConfigRequest.SerializeToString,
            draft_dot_protos_dot_ydb__dynamic__config__pb2.ResolveAllConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/ydb_cms.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v5.protos import ydb_operation_pb2 as protos_dot_ydb__operation__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14protos/ydb_cms.proto\x12\x07Ydb.Cms\x1a\x1aprotos/ydb_operation.proto\"0\n\x0cStorageUnits\x12\x11\n\tunit_kind\x18\x01 \x01(\t\x12\r\n\x05\x63ount\x18\x02 \x01(\x04\"Q\n\x12\x43omputationalUnits\x12\x11\n\tunit_kind\x18\x01 \x01(\t\x12\x19\n\x11\x61vailability_zone\x18\x02 \x01(\t\x12\r\n\x05\x63ount\x18\x03 \x01(\x04\"K\n\x1a\x41llocatedComputationalUnit\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\r\x12\x11\n\tunit_kind\x18\x03 \x01(\t\"s\n\tResources\x12,\n\rstorage_units\x18\x01 \x03(\x0b\x32\x15.Ydb.Cms.StorageUnits\x12\x38\n\x13\x63omputational_units\x18\x02 \x03(\x0b\x32\x1b.Ydb.Cms.ComputationalUnits\"3\n\x13ServerlessResources\x12\x1c\n\x14shared_database_path\x18\x01 \x01(\t\"j\n\x0f\x44\x61tabaseOptions\x12\x1a\n\x12\x64isable_tx_service\x18\x01 \x01(\x08\x12\"\n\x1a\x64isable_external_subdomain\x18\x02 \x01(\x08\x12\x17\n\x0fplan_resolution\x18\x03 \x01(\r\"\x9c\x01\n\x15SchemaOperationQuotas\x12G\n\x13leaky_bucket_quotas\x18\x01 \x03(\x0b\x32*.Ydb.Cms.SchemaOperationQuotas.LeakyBucket\x1a:\n\x0bLeakyBucket\x12\x13\n\x0b\x62ucket_size\x18\x01 \x01(\x01\x12\x16\n\x0e\x62ucket_seconds\x18\x02 \x01(\x04\"\xc0\x01\n\x0e\x44\x61tabaseQuotas\x12\x1c\n\x14\x64\x61ta_size_hard_quota\x18\x01 \x01(\x04\x12\x1c\n\x14\x64\x61ta_size_soft_quota\x18\x02 \x01(\x04\x12 \n\x18\x64\x61ta_stream_shards_quota\x18\x03 \x01(\x04\x12*\n\"data_stream_reserved_storage_quota\x18\x05 \x01(\x04\x12$\n\x1cttl_min_run_internal_seconds\x18\x04 \x01(\r\"\xb7\x04\n\x15\x43reateDatabaseRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\'\n\tresources\x18\x03 \x01(\x0b\x32\x12.Ydb.Cms.ResourcesH\x00\x12.\n\x10shared_resources\x18\x06 \x01(\x0b\x32\x12.Ydb.Cms.ResourcesH\x00\x12<\n\x14serverless_resources\x18\x07 \x01(\x0b\x32\x1c.Ydb.Cms.ServerlessResourcesH\x00\x12)\n\x07options\x18\x04 \x01(\x0b\x32\x18.Ydb.Cms.DatabaseOptions\x12\x42\n\nattributes\x18\x05 \x03(\x0b\x32..Ydb.Cms.CreateDatabaseRequest.AttributesEntry\x12?\n\x17schema_operation_quotas\x18\x08 \x01(\x0b\x32\x1e.Ydb.Cms.SchemaOperationQuotas\x12\x17\n\x0fidempotency_key\x18\t \x01(\t\x12\x30\n\x0f\x64\x61tabase_quotas\x18\n \x01(\x0b\x32\x17.Ydb.Cms.DatabaseQuotas\x1a\x31\n\x0f\x41ttributesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x10\n\x0eresources_kind\"F\n\x16\x43reateDatabaseResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"c\n\x18GetDatabaseStatusRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x39\n\x10operation_params\x18\x02 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\"I\n\x19GetDatabaseStatusResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\x85\x05\n\x17GetDatabaseStatusResult\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x35\n\x05state\x18\x02 \x01(\x0e\x32&.Ydb.Cms.GetDatabaseStatusResult.State\x12\x30\n\x12required_resources\x18\x03 \x01(\x0b\x32\x12.Ydb.Cms.ResourcesH\x00\x12\x37\n\x19required_shared_resources\x18\x07 \x01(\x0b\x32\x12.Ydb.Cms.ResourcesH\x00\x12<\n\x14serverless_resources\x18\x08 \x01(\x0b\x32\x1c.Ydb.Cms.ServerlessResourcesH\x00\x12/\n\x13\x61llocated_resources\x18\x04 \x01(\x0b\x32\x12.Ydb.Cms.Resources\x12\x41\n\x14registered_resources\x18\x05 \x03(\x0b\x32#.Ydb.Cms.AllocatedComputationalUnit\x12\x12\n\ngeneration\x18\x06 \x01(\x04\x12?\n\x17schema_operation_quotas\x18\t \x01(\x0b\x32\x1e.Ydb.Cms.SchemaOperationQuotas\x12\x30\n\x0f\x64\x61tabase_quotas\x18\n \x01(\x0b\x32\x17.Ydb.Cms.DatabaseQuotas\"o\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0c\n\x08\x43REATING\x10\x01\x12\x0b\n\x07RUNNING\x10\x02\x12\x0c\n\x08REMOVING\x10\x03\x12\x15\n\x11PENDING_RESOURCES\x10\x04\x12\x0f\n\x0b\x43ONFIGURING\x10\x05\x42\x10\n\x0eresources_kind\"\xdd\x05\n\x14\x41lterDatabaseRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12?\n\x1a\x63omputational_units_to_add\x18\x02 \x03(\x0b\x32\x1b.Ydb.Cms.ComputationalUnits\x12\x42\n\x1d\x63omputational_units_to_remove\x18\x03 \x03(\x0b\x32\x1b.Ydb.Cms.ComputationalUnits\x12\x33\n\x14storage_units_to_add\x18\x04 \x03(\x0b\x32\x15.Ydb.Cms.StorageUnits\x12L\n\x1f\x63omputational_units_to_register\x18\x05 \x03(\x0b\x32#.Ydb.Cms.AllocatedComputationalUnit\x12N\n!computational_units_to_deregister\x18\x06 \x03(\x0b\x32#.Ydb.Cms.AllocatedComputationalUnit\x12\x39\n\x10operation_params\x18\x07 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x12\n\ngeneration\x18\x08 \x01(\x04\x12?\n\x17schema_operation_quotas\x18\t \x01(\x0b\x32\x1e.Ydb.Cms.SchemaOperationQuotas\x12\x17\n\x0fidempotency_key\x18\n \x01(\t\x12\x30\n\x0f\x64\x61tabase_quotas\x18\x0b \x01(\x0b\x32\x17.Ydb.Cms.DatabaseQuotas\x12L\n\x10\x61lter_attributes\x18\x0c \x03(\x0b\x32\x32.Ydb.Cms.AlterDatabaseRequest.AlterAttributesEntry\x1a\x36\n\x14\x41lterAttributesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"E\n\x15\x41lterDatabaseResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"Q\n\x14ListDatabasesRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\"E\n\x15ListDatabasesResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"$\n\x13ListDatabasesResult\x12\r\n\x05paths\x18\x01 \x03(\t\"`\n\x15RemoveDatabaseRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x39\n\x10operation_params\x18\x02 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\"F\n\x16RemoveDatabaseResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\x92\x01\n\x16StorageUnitDescription\x12\x0c\n\x04kind\x18\x01 \x01(\t\x12;\n\x06labels\x18\x02 \x03(\x0b\x32+.Ydb.Cms.StorageUnitDescription.LabelsEntry\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x9c\x01\n\x1b\x41vailabilityZoneDescription\x12\x0c\n\x04name\x18\x01 \x01(\t\x12@\n\x06labels\x18\x02 \x03(\x0b\x32\x30.Ydb.Cms.AvailabilityZoneDescription.LabelsEntry\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xc2\x01\n\x1c\x43omputationalUnitDescription\x12\x0c\n\x04kind\x18\x01 \x01(\t\x12\x41\n\x06labels\x18\x02 \x03(\x0b\x32\x31.Ydb.Cms.ComputationalUnitDescription.LabelsEntry\x12\"\n\x1a\x61llowed_availability_zones\x18\x03 \x03(\t\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"[\n\x1e\x44\x65scribeDatabaseOptionsRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\"O\n\x1f\x44\x65scribeDatabaseOptionsResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\xdd\x01\n\x1d\x44\x65scribeDatabaseOptionsResult\x12\x36\n\rstorage_units\x18\x01 \x03(\x0b\x32\x1f.Ydb.Cms.StorageUnitDescription\x12@\n\x12\x61vailability_zones\x18\x02 \x03(\x0b\x32$.Ydb.Cms.AvailabilityZoneDescription\x12\x42\n\x13\x63omputational_units\x18\x03 \x03(\x0b\x32%.Ydb.Cms.ComputationalUnitDescriptionBO\n\x12tech.ydb.proto.cmsZ6github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Cms\xf8\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protos.ydb_cms_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\022tech.ydb.proto.cmsZ6github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Cms\370\001\001'
  _CREATEDATABASEREQUEST_ATTRIBUTESENTRY._options = None
  _CREATEDATABASEREQUEST_ATTRIBUTESENTRY._serialized_options = b'8\001'
  _ALTERDATABASEREQUEST_ALTERATTRIBUTESENTRY._options = None
  _ALTERDATABASEREQUEST_ALTERATTRIBUTESENTRY._serialized_options = b'8\001'
  _STORAGEUNITDESCRIPTION_LABELSENTRY._options = None
  _STORAGEUNITDESCRIPTION_LABELSENTRY._serialized_options = b'8\001'
  _AVAILABILITYZONEDESCRIPTION_LABELSENTRY._options = None
  _AVAILABILITYZONEDESCRIPTION_LABELSENTRY._serialized_options = b'8\001'
  _COMPUTATIONALUNITDESCRIPTION_LABELSENTRY._options = None
  _COMPUTATIONALUNITDESCRIPTION_LABELSENTRY._serialized_options = b'8\001'
  _STORAGEUNITS._serialized_start=61
  _STORAGEUNITS._serialized_end=109
  _COMPUTATIONALUNITS._serialized_start=111
  _COMPUTATIONALUNITS._serialized_end=192
  _ALLOCATEDCOMPUTATIONALUNIT._serialized_start=194
  _ALLOCATEDCOMPUTATIONALUNIT._serialized_end=269
  _RESOURCES._serialized_start=271
  _RESOURCES._serialized_end=386
  _SERVERLESSRESOURCES._serialized_start=388
  _SERVERLESSRESOURCES._serialized_end=439
  _DATABASEOPTIONS._serialized_start=441
  _DATABASEOPTIONS._serialized_end=547
  _SCHEMAOPERATIONQUOTAS._serialized_start=550
  _SCHEMAOPERATIONQUOTAS._serialized_end=706
  _SCHEMAOPERATIONQUOTAS_LEAKYBUCKET._serialized_start=648
  _SCHEMAOPERATIONQUOTAS_LEAKYBUCKET._serialized_end=706
  _DATABASEQUOTAS._serialized_start=709
  _DATABASEQUOTAS._serialized_end=901
  _CREATEDATABASEREQUEST._serialized_start=904
  _CREATEDATABASEREQUEST._serialized_end=1471
  _CREATEDATABASEREQUEST_ATTRIBUTESENTRY._serialized_start=1404
  _CREATEDATABASEREQUEST_ATTRIBUTESENTRY._serialized_end=1453
  _CREATEDATABASERESPONSE._serialized_start=1473
  _CREATEDATABASERESPONSE._serialized_end=1543
  _GETDATABASESTATUSREQUEST._serialized_start=1545
  _GETDATABASESTATUSREQUEST._serialized_end=1644
  _GETDATABASESTATUSRESPONSE._serialized_start=1646
  _GETDATABASESTATUSRESPONSE._serialized_end=1719
  _GETDATABASESTATUSRESULT._serialized_start=1722
  _GETDATABASESTATUSRESULT._serialized_end=2367
  _GETDATABASESTATUSRESULT_STATE._serialized_start=2238
  _GETDATABASESTATUSRESULT_STATE._serialized_end=2349
  _ALTERDATABASEREQUEST._serialized_start=2370
  _ALTERDATABASEREQUEST._serialized_end=3103
  _ALTERDATABASEREQUEST_ALTERATTRIBUTESENTRY._serialized_start=3049
  _ALTERDATABASEREQUEST_ALTERATTRIBUTESENTRY._serialized_end=3103
  _ALTERDATABASERESPONSE._serialized_start=3105
  _ALTERDATABASERESPONSE._serialized_end=3174
  _LISTDATABASESREQUEST._serialized_start=3176
  _LISTDATABASESREQUEST._serialized_end=3257
  _LISTDATABASESRESPONSE._serialized_start=3259
  _LISTDATABASESRESPONSE._serialized_end=3328
  _LISTDATABASESRESULT._serialized_start=3330
  _LISTDATABASESRESULT._serialized_end=3366
  _REMOVEDATABASEREQUEST._serialized_start=3368
  _REMOVEDATABASEREQUEST._serialized_end=3464
  _REMOVEDATABASERESPONSE._serialized_start=3466
  _REMOVEDATABASERESPONSE._serialized_end=3536
  _STORAGEUNITDESCRIPTION._serialized_start=3539
  _STORAGEUNITDESCRIPTION._serialized_end=3685
  _STORAGEUNITDESCRIPTION_LABELSENTRY._serialized_start=3640
  _STORAGEUNITDESCRIPTION_LABELSENTRY._serialized_end=3685
  _AVAILABILITYZONEDESCRIPTION._serialized_start=3688
  _AVAILABILITYZONEDESCRIPTION._serialized_end=3844
  _AVAILABILITYZONEDESCRIPTION_LABELSENTRY._serialized_start=3640
  _AVAILABILITYZONEDESCRIPTION_LABELSENTRY._serialized_end=3685
  _COMPUTATIONALUNITDESCRIPTION._serialized_start=3847
  _COMPUTATIONALUNITDESCRIPTION._serialized_end=4041
  _COMPUTATIONALUNITDESCRIPTION_LABELSENTRY._serialized_start=3640
  _COMPUTATIONALUNITDESCRIPTION_LABELSENTRY._serialized_end=3685
  _DESCRIBEDATABASEOPTIONSREQUEST._serialized_start=4043
  _DESCRIBEDATABASEOPTIONSREQUEST._serialized_end=4134
  _DESCRIBEDATABASEOPTIONSRESPONSE._serialized_start=4136
  _DESCRIBEDATABASEOPTIONSRESPONSE._serialized_end=4215
  _DESCRIBEDATABASEOPTIONSRESULT._serialized_start=4218
  _DESCRIBEDATABASEOPTIONSRESULT._serialized_end=4439
# @@protoc_insertion_point(module_scope)



import asyncio
from unittest.mock import MagicMock

from ydb.aio.driver import Driver
import pytest
import ydb


@pytest.mark.asyncio
async def test_async_call(endpoint, database):
    driver_config = ydb.DriverConfig(
        endpoint,
        database,
    )

    driver = Driver(driver_config=driver_config)

    await driver.scheme_client.make_directory("/local/lol")
    await driver.stop()


@pytest.mark.asyncio
async def test_gzip_compression(endpoint, database):
    driver_config = ydb.DriverConfig(
        endpoint,
        database,
        compression=ydb.RPCCompression.Gzip,
    )

    driver = Driver(driver_config=driver_config)

    await driver.scheme_client.make_directory(
        "/local/lol",
        settings=ydb.BaseRequestSettings().with_compression(ydb.RPCCompression.Deflate),
    )
    await driver.stop()


@pytest.mark.asyncio
async def test_other_credentials(endpoint, database):
    driver = Driver(endpoint=endpoint, database=database)

    await driver.scheme_client.make_directory("/local/lol")
    await driver.stop()


@pytest.mark.asyncio
async def test_session(endpoint, database):
    driver_config = ydb.DriverConfig(
        endpoint,
        database,
    )

    driver = Driver(driver_config=driver_config)

    await driver.wait(timeout=10)

    description = (
        ydb.TableDescription()
        .with_primary_keys("key1", "key2")
        .with_columns(
            ydb.Column("key1", ydb.OptionalType(ydb.PrimitiveType.Uint64)),
            ydb.Column("key2", ydb.OptionalType(ydb.PrimitiveType.Uint64)),
            ydb.Column("value", ydb.OptionalType(ydb.PrimitiveType.Utf8)),
        )
        .with_profile(
            ydb.TableProfile().with_partitioning_policy(
                ydb.PartitioningPolicy().with_explicit_partitions(
                    ydb.ExplicitPartitions(
                        (
                            ydb.KeyBound((100,)),
                            ydb.KeyBound((300, 100)),
                            ydb.KeyBound((400,)),
                        )
                    )
                )
            )
        )
    )

    session = await driver.table_client.session().create()
    await session.create_table(database + "/test_session", description)

    response = await session.describe_table(database + "/test_session")
    assert [c.name for c in response.columns] == ["key1", "key2", "value"]
    await driver.stop()


@pytest.mark.asyncio
async def test_raises_when_disconnect(endpoint, database, docker_project):

    driver_config = ydb.DriverConfig(
        endpoint,
        database,
    )

    driver = Driver(driver_config=driver_config)

    await driver.wait(timeout=10)

    async def restart_docker():
        docker_project.stop()

    coros = [driver.scheme_client.describe_path("/local") for i in range(100)]
    coros.append(restart_docker())

    with pytest.raises(ydb.ConnectionLost):
        await asyncio.gather(*coros, return_exceptions=False)

    docker_project.start()
    await driver.stop()


@pytest.mark.asyncio
async def test_disconnect_by_call(endpoint, database, docker_project):
    driver_config = ydb.DriverConfig(
        endpoint,
        database,
    )

    driver = Driver(driver_config=driver_config)

    await driver.wait(timeout=10)

    docker_project.stop()

    try:
        await driver.scheme_client.make_directory("/local/lol")
    except Exception:
        pass

    await asyncio.sleep(5)
    assert len(driver._store.connections) == 0
    docker_project.start()
    await driver.stop()


@pytest.mark.asyncio
async def test_good_discovery_interval(driver):
    await driver.wait(timeout=10)
    mock = MagicMock(return_value=0.1)
    driver._discovery._discovery_interval = mock
    driver._discovery._wake_up_event.set()
    await asyncio.sleep(1)
    assert mock.called



import json

import pytest

import ydb
from ydb import QueryExplainResultFormat
from ydb.aio.query.session import QuerySession
from ydb.connection import EndpointKey


def _check_session_not_ready(session: QuerySession):
    assert not session.is_active


def _check_session_ready(session: QuerySession):
    assert session.session_id is not None
    assert session.node_id is not None
    assert session.is_active
    assert not session.is_closed


class TestAsyncQuerySession:
    @pytest.mark.asyncio
    async def test_session_normal_lifecycle(self, session: QuerySession):
        _check_session_not_ready(session)

        await session.create()
        _check_session_ready(session)

        await session.delete()
        assert session.is_closed

    @pytest.mark.asyncio
    async def test_second_create_do_nothing(self, session: QuerySession):
        await session.create()
        _check_session_ready(session)

        session_id_before = session.session_id
        node_id_before = session.node_id

        await session.create()
        _check_session_ready(session)

        assert session.session_id == session_id_before
        assert session.node_id == node_id_before

    @pytest.mark.asyncio
    async def test_second_delete_do_nothing(self, session: QuerySession):
        await session.create()

        await session.delete()
        await session.delete()

    @pytest.mark.asyncio
    async def test_delete_before_create_is_noop(self, session: QuerySession):
        await session.delete()
        assert session.is_closed

    @pytest.mark.asyncio
    async def test_create_after_delete_not_possible(self, session: QuerySession):
        await session.create()
        await session.delete()
        with pytest.raises(RuntimeError):
            await session.create()

    def test_transaction_before_create_raises(self, session: QuerySession):
        with pytest.raises(RuntimeError):
            session.transaction()

    @pytest.mark.asyncio
    async def test_transaction_after_delete_raises(self, session: QuerySession):
        await session.create()

        await session.delete()

        with pytest.raises(RuntimeError):
            session.transaction()

    @pytest.mark.asyncio
    async def test_transaction_after_create_not_raises(self, session: QuerySession):
        await session.create()
        session.transaction()

    @pytest.mark.asyncio
    async def test_execute_before_create_raises(self, session: QuerySession):
        with pytest.raises(RuntimeError):
            await session.execute("select 1;")

    @pytest.mark.asyncio
    async def test_execute_after_delete_raises(self, session: QuerySession):
        await session.create()
        await session.delete()
        with pytest.raises(RuntimeError):
            await session.execute("select 1;")

    @pytest.mark.asyncio
    async def test_basic_execute(self, session: QuerySession):
        await session.create()
        it = await session.execute("select 1;")
        result_sets = [result_set async for result_set in it]

        assert len(result_sets) == 1
        assert len(result_sets[0].rows) == 1
        assert len(result_sets[0].columns) == 1
        assert list(result_sets[0].rows[0].values()) == [1]

    @pytest.mark.asyncio
    async def test_two_results(self, session: QuerySession):
        await session.create()
        res = []
        counter = 0

        async with await session.execute("select 1; select 2") as results:
            async for result_set in results:
                counter += 1
                if len(result_set.rows) > 0:
                    res.append(list(result_set.rows[0].values()))

        assert res == [[1], [2]]
        assert counter == 2

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("ydb_terminates_streams_with_unavailable")
    async def test_terminated_stream_raises_ydb_error(self, session: QuerySession):
        await session.create()

        with pytest.raises(ydb.Unavailable):
            async with await session.execute("select 1") as results:
                async for _ in results:
                    pass

    @pytest.mark.asyncio
    async def test_explain(self, pool: ydb.aio.query.QuerySessionPool):
        await pool.execute_with_retries("DROP TABLE IF EXISTS test_explain")
        await pool.execute_with_retries("CREATE TABLE test_explain (id Int64, PRIMARY KEY (id))")
        try:
            plan_fullscan = ""
            plan_lookup = ""

            async def callee(session: QuerySession):
                nonlocal plan_fullscan, plan_lookup

                plan = await session.explain("SELECT * FROM test_explain", result_format=QueryExplainResultFormat.STR)
                isinstance(plan, str)
                assert "FullScan" in plan

                plan_fullscan = await session.explain(
                    "SELECT * FROM test_explain", result_format=QueryExplainResultFormat.DICT
                )
                plan_lookup = await session.explain(
                    "SELECT * FROM test_explain WHERE id = $id",
                    {"$id": 1},
                    result_format=QueryExplainResultFormat.DICT,
                )

            await pool.retry_operation_async(callee)

            plan_fulltext_string = json.dumps(plan_fullscan)
            assert "FullScan" in plan_fulltext_string

            plan_lookup_string = json.dumps(plan_lookup)
            assert "Lookup" in plan_lookup_string
        finally:
            await pool.execute_with_retries("DROP TABLE test_explain")


class TestAsyncQuerySessionPreferredEndpoint:
    def test_endpoint_key_is_none_before_create(self, session: QuerySession):
        assert session._endpoint_key is None

    @pytest.mark.asyncio
    async def test_endpoint_key_is_set_after_create(self, session: QuerySession):
        await session.create()
        assert session.node_id is not None
        assert session._endpoint_key is not None
        assert isinstance(session._endpoint_key, EndpointKey)
        assert session._endpoint_key.node_id == session.node_id

    @pytest.mark.asyncio
    async def test_session_uses_preferred_endpoint_on_execute(self, session: QuerySession):
        await session.create()
        original_driver_call = session._driver

        calls = []

        async def mock_driver_call(*args, **kwargs):
            calls.append(kwargs)
            return await original_driver_call(*args, **kwargs)

        session._driver = mock_driver_call

        async with await session.execute("select 1;") as results:
            async for _ in results:
                pass

        assert len(calls) > 0
        assert "preferred_endpoint" in calls[0]
        assert calls[0]["preferred_endpoint"] is not None
        assert calls[0]["preferred_endpoint"].node_id == session.node_id

    @pytest.mark.asyncio
    async def test_session_uses_preferred_endpoint_on_delete(self, session: QuerySession):
        await session.create()
        original_driver_call = session._driver

        calls = []

        async def mock_driver_call(*args, **kwargs):
            calls.append(kwargs)
            return await original_driver_call(*args, **kwargs)

        session._driver = mock_driver_call

        await session.delete()

        assert len(calls) > 0
        assert "preferred_endpoint" in calls[0]
        assert calls[0]["preferred_endpoint"] is not None
        assert calls[0]["preferred_endpoint"].node_id == session.node_id

    @pytest.mark.asyncio
    async def test_transaction_uses_preferred_endpoint(self, session: QuerySession):
        await session.create()
        original_driver_call = session._driver

        calls = []

        async def mock_driver_call(*args, **kwargs):
            calls.append(kwargs)
            return await original_driver_call(*args, **kwargs)

        session._driver = mock_driver_call

        async with session.transaction() as tx:
            async with await tx.execute("select 1;") as results:
                async for _ in results:
                    pass

        execute_calls = [c for c in calls if "preferred_endpoint" in c]
        assert len(execute_calls) > 0
        for call in execute_calls:
            assert call["preferred_endpoint"] is not None
            assert call["preferred_endpoint"].node_id == session.node_id



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/annotations/sensitive.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='protos/annotations/sensitive.proto',
  package='Ydb',
  syntax='proto3',
  serialized_options=b'\n\016tech.ydb.protoZ2github.com/ydb-platform/ydb-go-genproto/protos/Ydb\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\"protos/annotations/sensitive.proto\x12\x03Ydb\x1a google/protobuf/descriptor.proto:2\n\tsensitive\x12\x1d.google.protobuf.FieldOptions\x18\xe7\xac\x05 \x01(\x08\x42G\n\x0etech.ydb.protoZ2github.com/ydb-platform/ydb-go-genproto/protos/Ydb\xf8\x01\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,])


SENSITIVE_FIELD_NUMBER = 87655
sensitive = _descriptor.FieldDescriptor(
  name='sensitive', full_name='Ydb.sensitive', index=0,
  number=87655, type=8, cpp_type=7, label=1,
  has_default_value=False, default_value=False,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)

DESCRIPTOR.extensions_by_name['sensitive'] = sensitive
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(sensitive)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import datetime

from protos.annotations import validation_pb2 as _validation_pb2
from protos import ydb_status_codes_pb2 as _ydb_status_codes_pb2
from protos import ydb_discovery_pb2 as _ydb_discovery_pb2
from protos import ydb_operation_pb2 as _ydb_operation_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ItemState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ITEM_STATE_UNSPECIFIED: _ClassVar[ItemState]
    ITEM_STATE_UP: _ClassVar[ItemState]
    ITEM_STATE_MAINTENANCE: _ClassVar[ItemState]
    ITEM_STATE_DOWN: _ClassVar[ItemState]

class AvailabilityMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    AVAILABILITY_MODE_UNSPECIFIED: _ClassVar[AvailabilityMode]
    AVAILABILITY_MODE_STRONG: _ClassVar[AvailabilityMode]
    AVAILABILITY_MODE_WEAK: _ClassVar[AvailabilityMode]
    AVAILABILITY_MODE_FORCE: _ClassVar[AvailabilityMode]
ITEM_STATE_UNSPECIFIED: ItemState
ITEM_STATE_UP: ItemState
ITEM_STATE_MAINTENANCE: ItemState
ITEM_STATE_DOWN: ItemState
AVAILABILITY_MODE_UNSPECIFIED: AvailabilityMode
AVAILABILITY_MODE_STRONG: AvailabilityMode
AVAILABILITY_MODE_WEAK: AvailabilityMode
AVAILABILITY_MODE_FORCE: AvailabilityMode

class Node(_message.Message):
    __slots__ = ("node_id", "host", "port", "location", "state", "storage", "dynamic", "start_time", "version")
    class StorageNode(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    class DynamicNode(_message.Message):
        __slots__ = ("tenant",)
        TENANT_FIELD_NUMBER: _ClassVar[int]
        tenant: str
        def __init__(self, tenant: _Optional[str] = ...) -> None: ...
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    HOST_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    STORAGE_FIELD_NUMBER: _ClassVar[int]
    DYNAMIC_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    node_id: int
    host: str
    port: int
    location: _ydb_discovery_pb2.NodeLocation
    state: ItemState
    storage: Node.StorageNode
    dynamic: Node.DynamicNode
    start_time: _timestamp_pb2.Timestamp
    version: str
    def __init__(self, node_id: _Optional[int] = ..., host: _Optional[str] = ..., port: _Optional[int] = ..., location: _Optional[_Union[_ydb_discovery_pb2.NodeLocation, _Mapping]] = ..., state: _Optional[_Union[ItemState, str]] = ..., storage: _Optional[_Union[Node.StorageNode, _Mapping]] = ..., dynamic: _Optional[_Union[Node.DynamicNode, _Mapping]] = ..., start_time: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., version: _Optional[str] = ...) -> None: ...

class ListClusterNodesRequest(_message.Message):
    __slots__ = ("operation_params",)
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ...) -> None: ...

class ListClusterNodesResult(_message.Message):
    __slots__ = ("nodes",)
    NODES_FIELD_NUMBER: _ClassVar[int]
    nodes: _containers.RepeatedCompositeFieldContainer[Node]
    def __init__(self, nodes: _Optional[_Iterable[_Union[Node, _Mapping]]] = ...) -> None: ...

class ListClusterNodesResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class MaintenanceTaskOptions(_message.Message):
    __slots__ = ("task_uid", "description", "availability_mode", "dry_run", "priority")
    TASK_UID_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    AVAILABILITY_MODE_FIELD_NUMBER: _ClassVar[int]
    DRY_RUN_FIELD_NUMBER: _ClassVar[int]
    PRIORITY_FIELD_NUMBER: _ClassVar[int]
    task_uid: str
    description: str
    availability_mode: AvailabilityMode
    dry_run: bool
    priority: int
    def __init__(self, task_uid: _Optional[str] = ..., description: _Optional[str] = ..., availability_mode: _Optional[_Union[AvailabilityMode, str]] = ..., dry_run: bool = ..., priority: _Optional[int] = ...) -> None: ...

class ActionScope(_message.Message):
    __slots__ = ("node_id", "host")
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    HOST_FIELD_NUMBER: _ClassVar[int]
    node_id: int
    host: str
    def __init__(self, node_id: _Optional[int] = ..., host: _Optional[str] = ...) -> None: ...

class LockAction(_message.Message):
    __slots__ = ("scope", "duration")
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    DURATION_FIELD_NUMBER: _ClassVar[int]
    scope: ActionScope
    duration: _duration_pb2.Duration
    def __init__(self, scope: _Optional[_Union[ActionScope, _Mapping]] = ..., duration: _Optional[_Union[datetime.timedelta, _duration_pb2.Duration, _Mapping]] = ...) -> None: ...

class Action(_message.Message):
    __slots__ = ("lock_action",)
    LOCK_ACTION_FIELD_NUMBER: _ClassVar[int]
    lock_action: LockAction
    def __init__(self, lock_action: _Optional[_Union[LockAction, _Mapping]] = ...) -> None: ...

class ActionGroup(_message.Message):
    __slots__ = ("actions",)
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    actions: _containers.RepeatedCompositeFieldContainer[Action]
    def __init__(self, actions: _Optional[_Iterable[_Union[Action, _Mapping]]] = ...) -> None: ...

class CreateMaintenanceTaskRequest(_message.Message):
    __slots__ = ("operation_params", "task_options", "action_groups")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    TASK_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    ACTION_GROUPS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    task_options: MaintenanceTaskOptions
    action_groups: _containers.RepeatedCompositeFieldContainer[ActionGroup]
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., task_options: _Optional[_Union[MaintenanceTaskOptions, _Mapping]] = ..., action_groups: _Optional[_Iterable[_Union[ActionGroup, _Mapping]]] = ...) -> None: ...

class RefreshMaintenanceTaskRequest(_message.Message):
    __slots__ = ("operation_params", "task_uid")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    TASK_UID_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    task_uid: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., task_uid: _Optional[str] = ...) -> None: ...

class ActionUid(_message.Message):
    __slots__ = ("task_uid", "group_id", "action_id")
    TASK_UID_FIELD_NUMBER: _ClassVar[int]
    GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    ACTION_ID_FIELD_NUMBER: _ClassVar[int]
    task_uid: str
    group_id: str
    action_id: str
    def __init__(self, task_uid: _Optional[str] = ..., group_id: _Optional[str] = ..., action_id: _Optional[str] = ...) -> None: ...

class ActionState(_message.Message):
    __slots__ = ("action", "action_uid", "status", "reason", "reason_details", "deadline")
    class ActionStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        ACTION_STATUS_UNSPECIFIED: _ClassVar[ActionState.ActionStatus]
        ACTION_STATUS_PENDING: _ClassVar[ActionState.ActionStatus]
        ACTION_STATUS_PERFORMED: _ClassVar[ActionState.ActionStatus]
    ACTION_STATUS_UNSPECIFIED: ActionState.ActionStatus
    ACTION_STATUS_PENDING: ActionState.ActionStatus
    ACTION_STATUS_PERFORMED: ActionState.ActionStatus
    class ActionReason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        ACTION_REASON_UNSPECIFIED: _ClassVar[ActionState.ActionReason]
        ACTION_REASON_OK: _ClassVar[ActionState.ActionReason]
        ACTION_REASON_TOO_MANY_UNAVAILABLE_VDISKS: _ClassVar[ActionState.ActionReason]
        ACTION_REASON_TOO_MANY_UNAVAILABLE_STATE_STORAGE_RINGS: _ClassVar[ActionState.ActionReason]
        ACTION_REASON_DISABLED_NODES_LIMIT_REACHED: _ClassVar[ActionState.ActionReason]
        ACTION_REASON_TENANT_DISABLED_NODES_LIMIT_REACHED: _ClassVar[ActionState.ActionReason]
        ACTION_REASON_WRONG_REQUEST: _ClassVar[ActionState.ActionReason]
        ACTION_REASON_SYS_TABLETS_NODE_LIMIT_REACHED: _ClassVar[ActionState.ActionReason]
        ACTION_REASON_GENERIC: _ClassVar[ActionState.ActionReason]
    ACTION_REASON_UNSPECIFIED: ActionState.ActionReason
    ACTION_REASON_OK: ActionState.ActionReason
    ACTION_REASON_TOO_MANY_UNAVAILABLE_VDISKS: ActionState.ActionReason
    ACTION_REASON_TOO_MANY_UNAVAILABLE_STATE_STORAGE_RINGS: ActionState.ActionReason
    ACTION_REASON_DISABLED_NODES_LIMIT_REACHED: ActionState.ActionReason
    ACTION_REASON_TENANT_DISABLED_NODES_LIMIT_REACHED: ActionState.ActionReason
    ACTION_REASON_WRONG_REQUEST: ActionState.ActionReason
    ACTION_REASON_SYS_TABLETS_NODE_LIMIT_REACHED: ActionState.ActionReason
    ACTION_REASON_GENERIC: ActionState.ActionReason
    ACTION_FIELD_NUMBER: _ClassVar[int]
    ACTION_UID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    REASON_DETAILS_FIELD_NUMBER: _ClassVar[int]
    DEADLINE_FIELD_NUMBER: _ClassVar[int]
    action: Action
    action_uid: ActionUid
    status: ActionState.ActionStatus
    reason: ActionState.ActionReason
    reason_details: str
    deadline: _timestamp_pb2.Timestamp
    def __init__(self, action: _Optional[_Union[Action, _Mapping]] = ..., action_uid: _Optional[_Union[ActionUid, _Mapping]] = ..., status: _Optional[_Union[ActionState.ActionStatus, str]] = ..., reason: _Optional[_Union[ActionState.ActionReason, str]] = ..., reason_details: _Optional[str] = ..., deadline: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class ActionGroupStates(_message.Message):
    __slots__ = ("action_states",)
    ACTION_STATES_FIELD_NUMBER: _ClassVar[int]
    action_states: _containers.RepeatedCompositeFieldContainer[ActionState]
    def __init__(self, action_states: _Optional[_Iterable[_Union[ActionState, _Mapping]]] = ...) -> None: ...

class MaintenanceTaskResult(_message.Message):
    __slots__ = ("task_uid", "action_group_states", "retry_after")
    TASK_UID_FIELD_NUMBER: _ClassVar[int]
    ACTION_GROUP_STATES_FIELD_NUMBER: _ClassVar[int]
    RETRY_AFTER_FIELD_NUMBER: _ClassVar[int]
    task_uid: str
    action_group_states: _containers.RepeatedCompositeFieldContainer[ActionGroupStates]
    retry_after: _timestamp_pb2.Timestamp
    def __init__(self, task_uid: _Optional[str] = ..., action_group_states: _Optional[_Iterable[_Union[ActionGroupStates, _Mapping]]] = ..., retry_after: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class MaintenanceTaskResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class GetMaintenanceTaskRequest(_message.Message):
    __slots__ = ("operation_params", "task_uid")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    TASK_UID_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    task_uid: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., task_uid: _Optional[str] = ...) -> None: ...

class GetMaintenanceTaskResult(_message.Message):
    __slots__ = ("task_options", "action_group_states")
    TASK_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    ACTION_GROUP_STATES_FIELD_NUMBER: _ClassVar[int]
    task_options: MaintenanceTaskOptions
    action_group_states: _containers.RepeatedCompositeFieldContainer[ActionGroupStates]
    def __init__(self, task_options: _Optional[_Union[MaintenanceTaskOptions, _Mapping]] = ..., action_group_states: _Optional[_Iterable[_Union[ActionGroupStates, _Mapping]]] = ...) -> None: ...

class GetMaintenanceTaskResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ListMaintenanceTasksRequest(_message.Message):
    __slots__ = ("operation_params", "user")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    USER_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    user: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., user: _Optional[str] = ...) -> None: ...

class ListMaintenanceTasksResult(_message.Message):
    __slots__ = ("tasks_uids",)
    TASKS_UIDS_FIELD_NUMBER: _ClassVar[int]
    tasks_uids: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, tasks_uids: _Optional[_Iterable[str]] = ...) -> None: ...

class ListMaintenanceTasksResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class DropMaintenanceTaskRequest(_message.Message):
    __slots__ = ("operation_params", "task_uid")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    TASK_UID_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    task_uid: str
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., task_uid: _Optional[str] = ...) -> None: ...

class ManageMaintenanceTaskResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class CompleteActionRequest(_message.Message):
    __slots__ = ("operation_params", "action_uids")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    ACTION_UIDS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    action_uids: _containers.RepeatedCompositeFieldContainer[ActionUid]
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., action_uids: _Optional[_Iterable[_Union[ActionUid, _Mapping]]] = ...) -> None: ...

class ManageActionResult(_message.Message):
    __slots__ = ("action_statuses",)
    class Status(_message.Message):
        __slots__ = ("action_uid", "status")
        ACTION_UID_FIELD_NUMBER: _ClassVar[int]
        STATUS_FIELD_NUMBER: _ClassVar[int]
        action_uid: ActionUid
        status: _ydb_status_codes_pb2.StatusIds.StatusCode
        def __init__(self, action_uid: _Optional[_Union[ActionUid, _Mapping]] = ..., status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ...) -> None: ...
    ACTION_STATUSES_FIELD_NUMBER: _ClassVar[int]
    action_statuses: _containers.RepeatedCompositeFieldContainer[ManageActionResult.Status]
    def __init__(self, action_statuses: _Optional[_Iterable[_Union[ManageActionResult.Status, _Mapping]]] = ...) -> None: ...

class ManageActionResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...



from __future__ import annotations

import asyncio
import logging
import typing
from concurrent.futures import Future
from typing import Union, List, Optional

from .._grpc.grpcwrapper.common_utils import SupportedDriverType
from .topic_writer import (
    PublicWriterSettings,
    PublicWriterInitInfo,
    PublicWriteResult,
    Message,
    TopicWriterClosedError,
)

from ..query.base import TxEvent

from .topic_writer_asyncio import (
    TxWriterAsyncIO,
    WriterAsyncIO,
)
from .._topic_common.common import (
    _get_shared_event_loop,
    TimeoutType,
    CallFromSyncToAsync,
)

if typing.TYPE_CHECKING:
    from ..query.transaction import BaseQueryTxContext

logger = logging.getLogger(__name__)


class WriterSync:
    _caller: CallFromSyncToAsync
    _async_writer: WriterAsyncIO
    _closed: bool
    _parent: typing.Any  # need for prevent close parent client by GC

    def __init__(
        self,
        driver: SupportedDriverType,
        settings: PublicWriterSettings,
        *,
        eventloop: Optional[asyncio.AbstractEventLoop] = None,
        _parent=None,
    ):

        self._closed = False

        if eventloop:
            loop = eventloop
        else:
            loop = _get_shared_event_loop()

        self._caller = CallFromSyncToAsync(loop)

        async def create_async_writer():
            return WriterAsyncIO(driver, settings)

        self._async_writer = self._caller.safe_call_with_result(create_async_writer(), None)
        self._parent = _parent

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.close()
        except BaseException:
            if exc_val is None:
                raise

    def __del__(self):
        if not self._closed:
            try:
                logger.debug("Topic writer was not closed properly. Consider using method close().")
                self.close(flush=False)
            except BaseException:
                logger.warning("Something went wrong during writer close in __del__")

    def close(self, *, flush: bool = True, timeout: TimeoutType = None):
        if self._closed:
            return

        logger.debug("Close topic writer")
        self._closed = True

        self._caller.safe_call_with_result(self._async_writer.close(flush=flush), timeout)

    def _check_closed(self):
        if self._closed:
            raise TopicWriterClosedError()

    def async_flush(self) -> Future:
        self._check_closed()

        return self._caller.unsafe_call_with_future(self._async_writer.flush())

    def flush(self, *, timeout=None):
        self._check_closed()

        logger.debug("flush writer")

        return self._caller.unsafe_call_with_result(self._async_writer.flush(), timeout)

    def async_wait_init(self) -> Future[PublicWriterInitInfo]:
        self._check_closed()

        logger.debug("wait writer init")

        return self._caller.unsafe_call_with_future(self._async_writer.wait_init())

    def wait_init(self, *, timeout: TimeoutType = None) -> PublicWriterInitInfo:
        self._check_closed()

        logger.debug("wait writer init")

        return self._caller.unsafe_call_with_result(self._async_writer.wait_init(), timeout)

    def write(
        self,
        messages: Union[Message, List[Message]],
        timeout: TimeoutType = None,
    ):
        self._check_closed()

        logger.debug(
            "write %s messages",
            len(messages) if isinstance(messages, list) else 1,
        )

        self._caller.safe_call_with_result(self._async_writer.write(messages), timeout)

    def async_write_with_ack(
        self,
        messages: Union[Message, List[Message]],
    ) -> Future[Union[PublicWriteResult, List[PublicWriteResult]]]:
        self._check_closed()

        return self._caller.unsafe_call_with_future(self._async_writer.write_with_ack(messages))

    def write_with_ack(
        self,
        messages: Union[Message, List[Message]],
        timeout: Union[float, None] = None,
    ) -> Union[PublicWriteResult, List[PublicWriteResult]]:
        self._check_closed()

        logger.debug(
            "write_with_ack %s messages",
            len(messages) if isinstance(messages, list) else 1,
        )

        return self._caller.unsafe_call_with_result(self._async_writer.write_with_ack(messages), timeout=timeout)


class TxWriterSync(WriterSync):
    def __init__(
        self,
        tx: "BaseQueryTxContext",
        driver: SupportedDriverType,
        settings: PublicWriterSettings,
        *,
        eventloop: Optional[asyncio.AbstractEventLoop] = None,
        _parent=None,
    ):

        self._closed = False

        if eventloop:
            loop = eventloop
        else:
            loop = _get_shared_event_loop()

        self._caller = CallFromSyncToAsync(loop)

        async def create_async_writer():
            return TxWriterAsyncIO(tx, driver, settings, _is_implicit=True)

        self._async_writer = self._caller.safe_call_with_result(create_async_writer(), None)
        self._parent = _parent

        tx._add_callback(TxEvent.BEFORE_COMMIT, self._on_before_commit, None)
        tx._add_callback(TxEvent.BEFORE_ROLLBACK, self._on_before_rollback, None)

    def _on_before_commit(self, tx: "BaseQueryTxContext"):
        self.close()

    def _on_before_rollback(self, tx: "BaseQueryTxContext"):
        self.close(flush=False)



# -*- coding: utf-8 -*-
import typing
import json
import abc
import os
import base64

try:
    import requests
except ImportError:
    requests = None

from ydb import credentials, tracing, issues
from .token_source import TokenSource, FixedTokenSource, JwtTokenSource


# method -> is HMAC
_supported_uppercase_jwt_algs = {
    "HS256": True,
    "HS384": True,
    "HS512": True,
    "RS256": False,
    "RS384": False,
    "RS512": False,
    "PS256": False,
    "PS384": False,
    "PS512": False,
    "ES256": False,
    "ES384": False,
    "ES512": False,
}


class Oauth2TokenExchangeCredentialsBase(abc.ABC):
    def __init__(
        self,
        token_endpoint: str,
        subject_token_source: typing.Optional[TokenSource] = None,
        actor_token_source: typing.Optional[TokenSource] = None,
        audience: typing.Union[typing.List[str], str, None] = None,
        scope: typing.Union[typing.List[str], str, None] = None,
        resource: typing.Union[typing.List[str], str, None] = None,
        grant_type: str = "urn:ietf:params:oauth:grant-type:token-exchange",
        requested_token_type: str = "urn:ietf:params:oauth:token-type:access_token",
    ):
        self._token_endpoint = token_endpoint
        self._subject_token_source = subject_token_source
        self._actor_token_source = actor_token_source
        self._audience = audience
        self._scope = scope
        self._resource = resource
        self._grant_type = grant_type
        self._requested_token_type = requested_token_type

        if not self._token_endpoint:
            raise Exception("Oauth2 token exchange: no token endpoint specified")

    def _process_response_status_code(self, content: str, status_code: int):
        if status_code == 403:
            raise issues.Unauthenticated(content)
        if status_code >= 500:
            raise issues.Unavailable(content)
        if status_code >= 400:
            raise issues.BadRequest(content)
        if status_code != 200:
            raise issues.Error(content)

    def _process_response_json(self, response_json):
        access_token = response_json["access_token"]
        expires_in = response_json["expires_in"]
        token_type = response_json["token_type"]
        scope = response_json.get("scope")
        if token_type.lower() != "bearer":
            raise Exception("Oauth2 token exchange: unsupported token type: {}".format(token_type))
        if expires_in <= 0:
            raise Exception("Oauth2 token exchange: incorrect expiration time: {}".format(expires_in))
        if scope and scope != self._get_scope_param():
            raise Exception(
                'Oauth2 token exchange: different scope. Expected: "{}", but got: "{}"'.format(
                    self._get_scope_param(), scope
                )
            )
        return {"access_token": "Bearer " + access_token, "expires_in": expires_in}

    def _get_scope_param(self) -> typing.Optional[str]:
        if self._scope is None:
            return None
        if isinstance(self._scope, str):
            return self._scope
        # list
        return " ".join(self._scope)

    def _make_token_request_params(self):
        params = {
            "grant_type": self._grant_type,
            "requested_token_type": self._requested_token_type,
        }
        if self._resource:
            params["resource"] = self._resource
        if self._audience:
            params["audience"] = self._audience
        scope = self._get_scope_param()
        if scope:
            params["scope"] = scope
        if self._subject_token_source is not None:
            t = self._subject_token_source.token()
            params["subject_token"] = t.token
            params["subject_token_type"] = t.token_type
        if self._actor_token_source is not None:
            t = self._actor_token_source.token()
            params["actor_token"] = t.token
            params["actor_token_type"] = t.token_type

        return params

    @classmethod
    def _jwt_token_source_from_config(cls, cfg_json):
        signing_method = cls._required_string_from_config(cfg_json, "alg")
        is_hmac = _supported_uppercase_jwt_algs.get(signing_method.upper(), None)
        if is_hmac is not None:  # we know this method => do uppercase
            signing_method = signing_method.upper()
        private_key = cls._required_string_from_config(cfg_json, "private-key")
        if is_hmac:  # decode from base64
            private_key = base64.b64decode(private_key + "===")  # to allow unpadded strings
        return JwtTokenSource(
            signing_method=signing_method,
            private_key=private_key,
            key_id=cls._string_with_default_from_config(cfg_json, "kid", None),
            issuer=cls._string_with_default_from_config(cfg_json, "iss", None),
            subject=cls._string_with_default_from_config(cfg_json, "sub", None),
            audience=cls._list_of_strings_or_single_from_config(cfg_json, "aud"),
            id=cls._string_with_default_from_config(cfg_json, "jti", None),
            token_ttl_seconds=cls._duration_seconds_from_config(cfg_json, "ttl", 3600),
        )

    @classmethod
    def _fixed_token_source_from_config(cls, cfg_json):
        return FixedTokenSource(
            cls._required_string_from_config(cfg_json, "token"),
            cls._required_string_from_config(cfg_json, "token-type"),
        )

    @classmethod
    def _token_source_from_config(cls, cfg_json, key_name):
        value = cfg_json.get(key_name, None)
        if value is None:
            return None
        if not isinstance(value, dict):
            raise Exception('Key "{}" is expected to be a json map'.format(key_name))

        source_type = cls._required_string_from_config(value, "type")
        if source_type.upper() == "FIXED":
            return cls._fixed_token_source_from_config(value)
        if source_type.upper() == "JWT":
            return cls._jwt_token_source_from_config(value)
        raise Exception('"{}": unknown token source type: "{}"'.format(key_name, source_type))

    @classmethod
    def _list_of_strings_or_single_from_config(cls, cfg_json, key_name):
        value = cfg_json.get(key_name, None)
        if value is None:
            return None
        if isinstance(value, list):
            for val in value:
                if not isinstance(val, str) or not val:
                    raise Exception(
                        'Key "{}" is expected to be a single string or list of nonempty strings'.format(key_name)
                    )
            return value
        else:
            if isinstance(value, str):
                return value
            raise Exception('Key "{}" is expected to be a single string or list of nonempty strings'.format(key_name))

    @classmethod
    def _required_string_from_config(cls, cfg_json, key_name):
        value = cfg_json.get(key_name, None)
        if value is None or not isinstance(value, str) or not value:
            raise Exception('Key "{}" is expected to be a nonempty string'.format(key_name))
        return value

    @classmethod
    def _string_with_default_from_config(cls, cfg_json, key_name, default_value):
        value = cfg_json.get(key_name, None)
        if value is None:
            return default_value
        if not isinstance(value, str):
            raise Exception('Key "{}" is expected to be a string'.format(key_name))
        return value

    @classmethod
    def _duration_seconds_from_config(cls, cfg_json, key_name, default_value):
        value = cfg_json.get(key_name, None)
        if value is None:
            return default_value
        if not isinstance(value, str):
            raise Exception('Key "{}" is expected to be a string'.format(key_name))
        multiplier = 1
        if value.endswith("s"):
            multiplier = 1
            value = value[:-1]
        elif value.endswith("m"):
            multiplier = 60
            value = value[:-1]
        elif value.endswith("h"):
            multiplier = 3600
            value = value[:-1]
        elif value.endswith("d"):
            multiplier = 3600 * 24
            value = value[:-1]
        elif value.endswith("ms"):
            multiplier = 1.0 / 1000
            value = value[:-2]
        elif value.endswith("us"):
            multiplier = 1.0 / 1000000
            value = value[:-2]
        elif value.endswith("ns"):
            multiplier = 1.0 / 1000000000
            value = value[:-2]
        f = float(value)
        if f < 0.0:
            raise Exception("{}: negative duration is not allowed".format(value))
        return int(f * multiplier)

    @classmethod
    def from_file(cls, cfg_file, iam_endpoint=None):
        """
        Create OAuth 2.0 token exchange protocol credentials from config file.

        https://www.rfc-editor.org/rfc/rfc8693
        Config file must be a valid json file

        Fields of json file
            grant-type:           [string] Grant type option (default: "urn:ietf:params:oauth:grant-type:token-exchange")
            res:                  [string | list of strings] Resource option (optional)
            aud:                  [string | list of strings] Audience option for token exchange request (optional)
            scope:                [string | list of strings] Scope option (optional)
            requested-token-type: [string] Requested token type option (default: "urn:ietf:params:oauth:token-type:access_token")
            subject-credentials:  [creds_json] Subject credentials options (optional)
            actor-credentials:    [creds_json] Actor credentials options (optional)
            token-endpoint:       [string] Token endpoint

        Fields of creds_json (JWT):
            type:                 [string] Token source type. Set JWT
            alg:                  [string] Algorithm for JWT signature.
                                        Supported algorithms can be listed
                                        with GetSupportedOauth2TokenExchangeJwtAlgorithms()
            private-key:          [string] (Private) key in PEM format (RSA, EC) or Base64 format (HMAC) for JWT signature
            kid:                  [string] Key id JWT standard claim (optional)
            iss:                  [string] Issuer JWT standard claim (optional)
            sub:                  [string] Subject JWT standard claim (optional)
            aud:                  [string | list of strings] Audience JWT standard claim (optional)
            jti:                  [string] JWT ID JWT standard claim (optional)
            ttl:                  [string] Token TTL (default: 1h)

        Fields of creds_json (FIXED):
            type:                 [string] Token source type. Set FIXED
            token:                [string] Token value
            token-type:           [string] Token type value. It will become
                                        subject_token_type/actor_token_type parameter
                                        in token exchange request (https://www.rfc-editor.org/rfc/rfc8693)
        """
        with open(os.path.expanduser(cfg_file), "r") as r:
            cfg = r.read()

        return cls.from_content(cfg, iam_endpoint=iam_endpoint)

    @classmethod
    def from_content(cls, cfg, iam_endpoint=None):
        try:
            cfg_json = json.loads(cfg)
        except Exception as ex:
            raise Exception("Failed to parse json config: {}".format(ex))

        if iam_endpoint is not None:
            token_endpoint = iam_endpoint
        else:
            token_endpoint = cfg_json.get("token-endpoint", "")

        subject_token_source = cls._token_source_from_config(cfg_json, "subject-credentials")
        actor_token_source = cls._token_source_from_config(cfg_json, "actor-credentials")
        audience = cls._list_of_strings_or_single_from_config(cfg_json, "aud")
        scope = cls._list_of_strings_or_single_from_config(cfg_json, "scope")
        resource = cls._list_of_strings_or_single_from_config(cfg_json, "res")
        grant_type = cls._string_with_default_from_config(
            cfg_json, "grant-type", "urn:ietf:params:oauth:grant-type:token-exchange"
        )
        requested_token_type = cls._string_with_default_from_config(
            cfg_json, "requested-token-type", "urn:ietf:params:oauth:token-type:access_token"
        )

        return cls(
            token_endpoint=token_endpoint,
            subject_token_source=subject_token_source,
            actor_token_source=actor_token_source,
            audience=audience,
            scope=scope,
            resource=resource,
            grant_type=grant_type,
            requested_token_type=requested_token_type,
        )


class Oauth2TokenExchangeCredentials(credentials.AbstractExpiringTokenCredentials, Oauth2TokenExchangeCredentialsBase):
    def __init__(
        self,
        token_endpoint: str,
        subject_token_source: typing.Optional[TokenSource] = None,
        actor_token_source: typing.Optional[TokenSource] = None,
        audience: typing.Union[typing.List[str], str, None] = None,
        scope: typing.Union[typing.List[str], str, None] = None,
        resource: typing.Union[typing.List[str], str, None] = None,
        grant_type: str = "urn:ietf:params:oauth:grant-type:token-exchange",
        requested_token_type: str = "urn:ietf:params:oauth:token-type:access_token",
        tracer=None,
    ):
        super(Oauth2TokenExchangeCredentials, self).__init__(tracer)
        Oauth2TokenExchangeCredentialsBase.__init__(
            self,
            token_endpoint,
            subject_token_source,
            actor_token_source,
            audience,
            scope,
            resource,
            grant_type,
            requested_token_type,
        )

    @tracing.with_trace()
    def _make_token_request(self):
        assert (
            requests is not None
        ), "Install requests library to use Oauth2TokenExchangeCredentials credentials provider"

        params = self._make_token_request_params()
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        response = requests.post(self._token_endpoint, data=params, headers=headers)
        self._process_response_status_code(response.content, response.status_code)
        response_json = json.loads(response.content)
        return self._process_response_json(response_json)
